"""Threading and multiprocessing functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03a_parallel.ipynb.

# %% auto #0
__all__ = ['threaded', 'startthread', 'startproc', 'parallelable', 'ThreadPoolExecutor', 'ProcessPoolExecutor', 'parallel',
           'parallel_async', 'bg_task', 'run_procs', 'parallel_gen']

# %% ../nbs/03a_parallel.ipynb #569d18c6
from .imports import *
from .basics import *
from .foundation import *
from .meta import *
from .xtras import *
from functools import wraps

import concurrent.futures,time,asyncio
from multiprocessing import Process,Queue,Manager,set_start_method,get_all_start_methods,get_context
from threading import Thread
try:
    if sys.platform == 'darwin' and IN_NOTEBOOK: set_start_method("fork")
except: pass

# %% ../nbs/03a_parallel.ipynb #21e71104
def threaded(process=False):
    "Run `f` in a `Thread` (or `Process` if `process=True`), and returns it"
    def _r(f):
        def g(_obj_td, *args, **kwargs):
            res = f(*args, **kwargs)
            _obj_td.result = res
        @wraps(f)
        def _f(*args, **kwargs):
            Proc = get_context('fork').Process if sys.platform == 'darwin' else Process
            res = (Thread,Proc)[process](target=g, args=args, kwargs=kwargs)
            res._args = (res,)+res._args
            res.start()
            return res
        return _f
    if callable(process):
        o = process
        process = False
        return _r(o)
    return _r

# %% ../nbs/03a_parallel.ipynb #dbccb558
def startthread(f):
    "Like `threaded`, but start thread immediately"
    return threaded(f)()

# %% ../nbs/03a_parallel.ipynb #a32d66c4
def startproc(f):
    "Like `threaded(True)`, but start Process immediately"
    return threaded(True)(f)()

# %% ../nbs/03a_parallel.ipynb #44d4651b
def _call(lock, pause, n, g, item):
    l = False
    if pause:
        try:
            l = lock.acquire(timeout=pause*(n+2))
            time.sleep(pause)
        finally:
            if l: lock.release()
    return g(item)

# %% ../nbs/03a_parallel.ipynb #63a3920f
def parallelable(param_name, num_workers, f=None):
    f_in_main = f == None or sys.modules[f.__module__].__name__ == "__main__"
    if sys.platform == "win32" and IN_NOTEBOOK and num_workers > 0 and f_in_main:
        print("Due to IPython and Windows limitation, python multiprocessing isn't available now.")
        print(f"So `{param_name}` has to be changed to 0 to avoid getting stuck")
        return False
    return True

# %% ../nbs/03a_parallel.ipynb #e74836fc
class ThreadPoolExecutor(concurrent.futures.ThreadPoolExecutor):
    "Same as Python's ThreadPoolExecutor, except can pass `max_workers==0` for serial execution"
    def __init__(self, max_workers=defaults.cpus, on_exc=print, pause=0, **kwargs):
        if max_workers is None: max_workers=defaults.cpus
        store_attr()
        self.not_parallel = max_workers==0
        if self.not_parallel: max_workers=1
        super().__init__(max_workers, **kwargs)

    def map(self, f, items, *args, timeout=None, chunksize=1, **kwargs):
        if self.not_parallel == False: self.lock = Manager().Lock()
        g = partial(f, *args, **kwargs)
        if self.not_parallel: return map(g, items)
        _g = partial(_call, self.lock, self.pause, self.max_workers, g)
        try: return super().map(_g, items, timeout=timeout, chunksize=chunksize)
        except Exception as e: self.on_exc(e)

# %% ../nbs/03a_parallel.ipynb #0046c557
@delegates()
class ProcessPoolExecutor(concurrent.futures.ProcessPoolExecutor):
    "Same as Python's ProcessPoolExecutor, except can pass `max_workers==0` for serial execution"
    def __init__(self, max_workers=defaults.cpus, on_exc=print, pause=0, **kwargs):
        if max_workers is None: max_workers=defaults.cpus
        store_attr()
        self.not_parallel = max_workers==0
        if self.not_parallel: max_workers=1
        super().__init__(max_workers, **kwargs)

    def map(self, f, items, *args, timeout=None, chunksize=1, **kwargs):
        if not parallelable('max_workers', self.max_workers, f): self.max_workers = 0
        self.not_parallel = self.max_workers==0
        if self.not_parallel: self.max_workers=1

        if self.not_parallel == False: self.lock = Manager().Lock()
        g = partial(f, *args, **kwargs)
        if self.not_parallel: return map(g, items)
        _g = partial(_call, self.lock, self.pause, self.max_workers, g)
        try: return super().map(_g, items, timeout=timeout, chunksize=chunksize)
        except Exception as e: self.on_exc(e)

# %% ../nbs/03a_parallel.ipynb #f5327cb3
try: from fastprogress import progress_bar
except: progress_bar = None

# %% ../nbs/03a_parallel.ipynb #529e1bb1
def parallel(f, items, *args, n_workers=defaults.cpus, total=None, progress=None, pause=0,
             method=None, threadpool=False, timeout=None, chunksize=1, **kwargs):
    "Applies `func` in parallel to `items`, using `n_workers`"
    kwpool = {}
    if threadpool: pool = ThreadPoolExecutor
    else:
        if not method and sys.platform == 'darwin': method='fork'
        if method: kwpool['mp_context'] = get_context(method)
        pool = ProcessPoolExecutor
    with pool(n_workers, pause=pause, **kwpool) as ex:
        r = ex.map(f,items, *args, timeout=timeout, chunksize=chunksize, **kwargs)
        if progress and progress_bar:
            if total is None: total = len(items)
            r = progress_bar(r, total=total, leave=False)
        return L(r)

# %% ../nbs/03a_parallel.ipynb #343d0191
def _add_one(x, a=1):
    # this import is necessary for multiprocessing in notebook on windows
    import random
    time.sleep(random.random()/80)
    return x+a

# %% ../nbs/03a_parallel.ipynb #87a80e04
async def parallel_async(f, items, *args, n_workers=16, pause=0,
                         timeout=None, chunksize=1, on_exc=print, cancel_on_error=False, **kwargs):
    "Applies `f` to `items` in parallel using asyncio and a semaphore to limit concurrency."
    semaphore = asyncio.Semaphore(n_workers)
    async def limited_task(i, item):
        coro = f(item, *args, **kwargs) if asyncio.iscoroutinefunction(f) else asyncio.to_thread(f, item, *args, **kwargs)
        if pause: await asyncio.sleep(i * pause)
        async with semaphore:
            return await asyncio.wait_for(coro, timeout) if timeout else await coro
    if cancel_on_error:
        async with asyncio.TaskGroup() as tg: tasks = [tg.create_task(limited_task(i, item)) for i,item in enumerate(items)]
        return [t.result() for t in tasks]
    return await asyncio.gather(*[limited_task(i, item) for i,item in enumerate(items)], return_exceptions=True)

# %% ../nbs/03a_parallel.ipynb #6748aa27
def bg_task(coro):
    "Like `asyncio.create_task` but logs exceptions for fire-and-forget tasks"
    import traceback,asyncio
    def _done(t):
        if not t.cancelled() and (exc := t.exception()): traceback.print_exception(exc)
    task = asyncio.create_task(coro)
    task.add_done_callback(_done)
    return task

# %% ../nbs/03a_parallel.ipynb #da364301
def run_procs(f, f_done, args):
    "Call `f` for each item in `args` in parallel, yielding `f_done`"
    Proc = get_context('fork').Process if sys.platform == 'darwin' else Process
    processes = L(args).map(Proc, args=arg0, target=f)
    for o in processes: o.start()
    yield from f_done()
    processes.map(Self.join())

# %% ../nbs/03a_parallel.ipynb #98698c9d
def _f_pg(obj, queue, batch, start_idx):
    for i,b in enumerate(obj(batch)): queue.put((start_idx+i,b))

def _done_pg(queue, items): return (queue.get() for _ in items)

# %% ../nbs/03a_parallel.ipynb #1122caee
def parallel_gen(cls, items, n_workers=defaults.cpus, progress=True, **kwargs):
    "Instantiate `cls` in `n_workers` procs & call each on a subset of `items` in parallel."
    if not parallelable('n_workers', n_workers): n_workers = 0
    if n_workers==0:
        yield from enumerate(list(cls(**kwargs)(items)))
        return
    batches = L(chunked(items, n_chunks=n_workers))
    idx = L(itertools.accumulate(0 + batches.map(len)))
    queue = Queue()
    if progress_bar and progress: items = progress_bar(items, leave=False)
    f=partial(_f_pg, cls(**kwargs), queue)
    done=partial(_done_pg, queue, items)
    yield from run_procs(f, done, L(batches,idx).zip())



"""Utility functions used in the fastai library"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_xtras.ipynb.

# %% ../nbs/03_xtras.ipynb #ecd054a6
from __future__ import annotations

# %% auto #0
__all__ = ['spark_chars', 'UNSET', 'walk', 'exttypes', 'globtastic', 'pglob', 'maybe_open', 'mkdir', 'image_size', 'img_bytes',
           'detect_mime', 'bunzip', 'loads', 'loads_multi', 'dumps', 'untar_dir', 'repo_details', 'shell', 'ssh',
           'rsync_multi', 'run', 'open_file', 'save_pickle', 'load_pickle', 'parse_env', 'expand_wildcards',
           'atomic_save', 'dict2obj', 'obj2dict', 'repr_dict', 'is_listy', 'mapped', 'IterLen', 'ReindexCollection',
           'SaveReturn', 'trim_wraps', 'save_iter', 'asave_iter', 'frontmatter', 'clean_cli_output', 'unqid',
           'rtoken_hex', 'friendly_name', 'n_friendly_names', 'exec_eval', 'get_source_link', 'sparkline',
           'modify_exception', 'round_multiple', 'set_num_threads', 'join_path_file', 'autostart', 'EventTimer',
           'stringfmt_names', 'PartialFormatter', 'partial_format', 'truncstr', 'utc2local', 'local2utc', 'trace',
           'modified_env', 'ContextManagers', 'shufflish', 'console_help', 'hl_md', 'type2str', 'dataclass_src',
           'Unset', 'nullable_dc', 'make_nullable', 'flexiclass', 'asdict', 'vars_pub', 'is_typeddict', 'is_namedtuple',
           'CachedIter', 'CachedAwaitable', 'reawaitable', 'is_async_callable', 'maybe_await', 'noopa', 'flexicache',
           'time_policy', 'mtime_policy', 'timed_cache']

# %% ../nbs/03_xtras.ipynb #3401d507
from .imports import *
from .foundation import *
from .basics import *
from .ansi import strip_ansi

from importlib import import_module
from functools import wraps
import string,time,dataclasses
from enum import Enum
from contextlib import contextmanager,ExitStack
from datetime import datetime, timezone
from time import sleep,time,perf_counter
from os.path import getmtime
from dataclasses import dataclass, field, fields, is_dataclass, MISSING, make_dataclass

# %% ../nbs/03_xtras.ipynb #f64c2b09
def walk(
    path:Path|str, # path to start searching
    symlinks:bool=True, # follow symlinks?
    keep_file:callable=ret_true, # function that returns True for wanted files
    keep_folder:callable=ret_true, # function that returns True for folders to enter
    skip_folder:callable=ret_false, # function that returns True for folders to skip
    func:callable=os.path.join, # function to apply to each matched file
    ret_folders:bool=False, # return folders, not just files
    sort:bool=True # sort files by name within each folder
):
    "Generator version of `os.walk`, using functions to filter files and folders"
    from copy import copy
    for root,dirs,files in os.walk(path, followlinks=symlinks):
        if keep_folder(root,''):
            if ret_folders: yield func(root, '')
            if sort: files = sorted(files)
            yield from (func(root, name) for name in files if keep_file(root,name))
        for name in copy(dirs):
            if skip_folder(root,name): dirs.remove(name)

# %% ../nbs/03_xtras.ipynb #12a01ff4
_exttypes = dict(
    py=['ipynb', 'py'],
    js=['js', 'jsx', 'mjs', 'cjs'],
    java=['java', 'kt', 'scala', 'gradle'],
    c=['c', 'h'],
    cpp=['cpp', 'hpp', 'cc', 'cxx'],
    rb=['rb', 'rake'],
    r=['r', 'R'],
    ex=['ex', 'exs', 'erl', 'hrl'],
    sh=['sh', 'bash'],
    web=['html', 'css'],
    doc=['md', 'rst'],
    cfg=['json', 'yaml', 'yml', 'toml', 'xml'],
)

def exttypes(types):
    """Get exts for comma-separated or list `typ`; if not found in list, return list with just `types`.
    Supported: py, js, java, c, cpp, rb, r, ex, sh, web, doc, cfg"""
    typs = L(types).flatmap(Self.split(','))
    return typs.flatmap(lambda x: _exttypes.get(x, [x]))

# %% ../nbs/03_xtras.ipynb #4fef7fef
def globtastic(
    path:Path|str='.', # path to start searching
    recursive:bool=True, # search subfolders
    symlinks:bool=True, # follow symlinks?
    file_glob:str=None, # Only include files matching glob
    file_re:str=None, # Only include files matching regex
    folder_re:str=None, # Only enter folders matching regex
    skip_file_glob:str=None, # Skip files matching glob
    skip_file_re:str=None, # Skip files matching regex
    skip_folder_re:str=None, # Skip folders matching regex,
    func:callable=os.path.join, # function to apply to each matched file
    ret_folders:bool=False, # return folders, not just files
    sort:bool=True, # sort files by name within each folder
    types:str|list=None,  # list or comma-separated str of ext types from: py, js, java, c, cpp, rb, r, ex, sh, web, doc, cfg
    exts:str|list=None, # list or comma-separated str of exts to include
)->L: # Paths to matched files
    "A more powerful `glob`, including regex matches, symlink handling, and skip parameters"
    from fnmatch import fnmatch
    path = Path(path)
    if path.is_file(): return L([path])
    if not recursive: skip_folder_re='.'
    exts = L(exts).flatmap(Self.split(','))
    exts += exttypes(types)
    if exts:
        exts = [e if e.startswith('.') else f'.{e}' for e in listify(exts)]
        file_re = f"({'|'.join(re.escape(e) for e in exts)})$"
    file_re,folder_re = compile_re(file_re),compile_re(folder_re)
    skip_file_re,skip_folder_re = compile_re(skip_file_re),compile_re(skip_folder_re)
    def _keep_file(root, name):
        return (not file_glob or fnmatch(name, file_glob)) and (
                not file_re or file_re.search(name)) and (
                not skip_file_glob or not fnmatch(name, skip_file_glob)) and (
                not skip_file_re or not skip_file_re.search(name))
    def _keep_folder(root, name): return not folder_re or folder_re.search(os.path.join(root,name))
    def _skip_folder(root, name): return skip_folder_re and skip_folder_re.search(name)
    return L(walk(path, symlinks=symlinks, keep_file=_keep_file, keep_folder=_keep_folder, skip_folder=_skip_folder,
                  func=func, ret_folders=ret_folders, sort=sort))

# %% ../nbs/03_xtras.ipynb #827b4914
@fdelegates(globtastic)
def pglob(
    path:Path|str='.', # path to start searching
    func:callable=Path, # function to apply to each matched file
    **kwargs
)->L: # Paths to matched files
    "Shortcut for `globtastic(..., call=Path)`"
    return globtastic(path, func=func, **kwargs)

# %% ../nbs/03_xtras.ipynb #d514da68
@contextmanager
def maybe_open(f, mode='r', **kwargs):
    "Context manager: open `f` if it is a path (and close on exit)"
    if isinstance(f, (str,os.PathLike)):
        with open(f, mode, **kwargs) as f: yield f
    else: yield f

# %% ../nbs/03_xtras.ipynb #63d86576
def mkdir(path, exist_ok=False, parents=False, overwrite=False, **kwargs):
    "Creates and returns a directory defined by `path`, optionally removing previous existing directory if `overwrite` is `True`"
    import shutil
    path = Path(path)
    if path.exists() and overwrite: shutil.rmtree(path)
    path.mkdir(exist_ok=exist_ok, parents=parents, **kwargs)
    return path

# %% ../nbs/03_xtras.ipynb #b33313ff
def image_size(fn):
    "Tuple of (w,h) for png, gif, or jpg; `None` otherwise"
    from fastcore import imghdr
    import struct
    def _jpg_size(f):
        size,ftype = 2,0
        while not 0xc0 <= ftype <= 0xcf:
            f.seek(size, 1)
            byte = f.read(1)
            while ord(byte) == 0xff: byte = f.read(1)
            ftype = ord(byte)
            size = struct.unpack('>H', f.read(2))[0] - 2
        f.seek(1, 1)  # `precision'
        h,w = struct.unpack('>HH', f.read(4))
        return w,h

    def _gif_size(f): return struct.unpack('<HH', head[6:10])

    def _png_size(f):
        assert struct.unpack('>i', head[4:8])[0]==0x0d0a1a0a
        return struct.unpack('>ii', head[16:24])
    d = dict(png=_png_size, gif=_gif_size, jpeg=_jpg_size)
    with maybe_open(fn, 'rb') as f: return d[imghdr.what(f)](f)

# %% ../nbs/03_xtras.ipynb #d3aa6edf
def img_bytes(img, fmt='PNG'):
    # Convert PIL `img` to bytes in format `fmt`
    from io import BytesIO
    buf=BytesIO()
    img.save(buf, format=fmt)
    return buf.getvalue()

# %% ../nbs/03_xtras.ipynb #e5801198
_sigs = {
    (b'%PDF', 0): 'application/pdf',
    (b'RIFF', 0): lambda d: 'audio/wav' if d[8:12]==b'WAVE' else 'video/avi' if d[8:12]==b'AVI ' else None,
    (b'ID3', 0): 'audio/mp3',
    (b'\xff\xfb', 0): 'audio/mp3',
    (b'\xff\xf3', 0): 'audio/mp3',
    (b'FORM', 0): lambda d: 'audio/aiff' if d[8:12]==b'AIFF' else None,
    (b'OggS', 0): 'audio/ogg',
    (b'fLaC', 0): 'audio/flac',
    (b'ftyp', 4): lambda d: 'video/3gpp' if d[8:11]==b'3gp' else 'video/mp4',
    (b'\x1a\x45\xdf', 0): 'video/webm',
    (b'FLV', 0): 'video/x-flv',
    (b'\x30\x26\xb2\x75', 0): 'video/wmv',
    (b'\x00\x00\x01\xb3', 0): 'video/mpeg',
}

def detect_mime(data):
    "Get the MIME type for bytes `data`, covering common PDF, audio, video, and image types"
    if not isinstance(data,bytes): return None
    import mimetypes
    from fastcore import imghdr
    for (sig,pos),mime in _sigs.items():
        if data[pos:pos+len(sig)]==sig: return mime(data) if callable(mime) else mime
    return mimetypes.types_map.get(f'.{imghdr.what(None, h=data)}')

# %% ../nbs/03_xtras.ipynb #f856746a
def bunzip(fn):
    "bunzip `fn`, raising exception if output already exists"
    fn = Path(fn)
    assert fn.exists(), f"{fn} doesn't exist"
    out_fn = fn.with_suffix('')
    assert not out_fn.exists(), f"{out_fn} already exists"
    import bz2
    with bz2.BZ2File(fn, 'rb') as src, out_fn.open('wb') as dst:
        for d in iter(lambda: src.read(1024*1024), b''): dst.write(d)

# %% ../nbs/03_xtras.ipynb #0a78c637
def loads(s, **kw):
    "Same as `json.loads`, but handles `None`"
    if not s: return {}
    try: import ujson as json
    except ModuleNotFoundError: import json
    return json.loads(s, **kw)

# %% ../nbs/03_xtras.ipynb #131170f7
def loads_multi(s:str):
    "Generator of >=0 decoded json dicts, possibly with non-json ignored text at start and end"
    import json
    _dec = json.JSONDecoder()
    while s.find('{')>=0:
        s = s[s.find('{'):]
        obj,pos = _dec.raw_decode(s)
        if not pos: raise ValueError(f'no JSON object found at {pos}')
        yield obj
        s = s[pos:]

# %% ../nbs/03_xtras.ipynb #ed688ac4
def dumps(obj, **kw):
    "Same as `json.dumps`, but uses `ujson` if available"
    try: import ujson as json
    except ModuleNotFoundError: import json
    else: kw['escape_forward_slashes']=False
    return json.dumps(obj, **kw)

# %% ../nbs/03_xtras.ipynb #ac84300c
def _unpack(fname, out):
    import shutil
    shutil.unpack_archive(str(fname), str(out))
    ls = out.ls()
    return ls[0] if len(ls) == 1 else out

# %% ../nbs/03_xtras.ipynb #6f1e90e9
def untar_dir(fname, dest, rename=False, overwrite=False):
    "untar `file` into `dest`, creating a directory if the root contains more than one item"
    import tempfile,shutil
    dest = Path(dest)
    with tempfile.TemporaryDirectory() as d:
        out = Path(d)/remove_suffix(Path(fname).stem, '.tar')
        out.mkdir()
        if rename: dest = dest/out.name
        else:
            src = _unpack(fname, out)
            dest = dest/src.name
        if dest.exists():
            if overwrite: shutil.rmtree(dest) if dest.is_dir() else dest.unlink()
            else: return dest
        if rename: src = _unpack(fname, out)
        shutil.move(str(src), dest)
        return dest

# %% ../nbs/03_xtras.ipynb #ab91ee87
def repo_details(url):
    "Tuple of `owner,name` from ssh or https git repo `url`"
    res = remove_suffix(url.strip(), '.git')
    res = res.split(':')[-1]
    return res.split('/')[-2:]

# %% ../nbs/03_xtras.ipynb #ba70e535
def shell(*args, **kwargs):
    "Shortcut for `subprocess.run(shell=True)`"
    import subprocess
    return subprocess.run(*args, shell=True, **kwargs)

# %% ../nbs/03_xtras.ipynb #f62134a5
def ssh(host, args='', user='ubuntu', sock=None):
    "Run SSH command with given arguments"
    sock_opts = f'-S {sock}' if sock else ''
    return shell(f'ssh {sock_opts} {args} {user}@{host}')

# %% ../nbs/03_xtras.ipynb #cb24780e
def rsync_multi(ip, files, user='ubuntu', persist='5m'):
    "Transfer multiple files with rename using persistent SSH connection"
    sock = f'/tmp/ssh-{ip}-{user}'
    ssh(ip, f'-o ControlMaster=auto -o ControlPersist={persist} -N -f', user, sock)
    for src,dst in files: shell(f'rsync -az -e "ssh -S {sock}" {src} {user}@{ip}:{dst}')

# %% ../nbs/03_xtras.ipynb #dda52646
def run(cmd, *rest, same_in_win=False, ignore_ex=False, as_bytes=False, stderr=True):
    "Pass `cmd` (splitting with `shlex` if string) to `subprocess.run`; return `stdout`; raise `IOError` if fails"
    # Even the command is same on Windows, we have to add `cmd /c `"
    import subprocess
    if rest:
        if sys.platform == 'win32' and same_in_win:
            cmd = ('cmd', '/c', cmd, *rest)
        else:
            cmd = (cmd,)+rest
    elif isinstance(cmd, str):
        if sys.platform == 'win32' and same_in_win: cmd = 'cmd /c ' + cmd
        import shlex
        cmd = shlex.split(cmd)
    elif isinstance(cmd, list):
        if sys.platform == 'win32' and same_in_win: cmd = ['cmd', '/c'] + cmd
    res = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout = res.stdout
    if stderr and res.stderr: stdout += b' ;; ' + res.stderr
    if not as_bytes: stdout = stdout.decode().strip()
    if ignore_ex: return (res.returncode, stdout)
    if res.returncode: raise IOError((res.stdout + b' ;; ' + res.stderr).decode().strip())
    return stdout

# %% ../nbs/03_xtras.ipynb #5097b830
def open_file(fn, mode='r', **kwargs):
    "Open a file, with optional compression if gz or bz2 suffix"
    if isinstance(fn, io.IOBase): return fn
    import bz2,gzip,zipfile
    fn = Path(fn)
    if   fn.suffix=='.bz2': return bz2.BZ2File(fn, mode, **kwargs)
    elif fn.suffix=='.gz' : return gzip.GzipFile(fn, mode, **kwargs)
    elif fn.suffix=='.zip': return zipfile.ZipFile(fn, mode, **kwargs)
    else: return open(fn,mode, **kwargs)

# %% ../nbs/03_xtras.ipynb #a9fc7a40
def save_pickle(fn, o):
    "Save a pickle file, to a file name or opened file"
    import pickle
    with open_file(fn, 'wb') as f: pickle.dump(o, f)

# %% ../nbs/03_xtras.ipynb #5e9bc30e
def load_pickle(fn):
    "Load a pickle file from a file name or opened file"
    import pickle
    with open_file(fn, 'rb') as f: return pickle.load(f)

# %% ../nbs/03_xtras.ipynb #02029d02
def parse_env(s:str=None, fn:Union[str,Path]=None) -> dict:
    "Parse a shell-style environment string or file"
    assert bool(s)^bool(fn), "Must pass exactly one of `s` or `fn`"
    if fn: s = Path(fn).read_text()
    def _f(line):
        m = re.match(r'^\s*(?:export\s+)?(\w+)\s*=\s*(["\']?)(.*?)(\2)\s*(?:#.*)?$', line).groups()
        return m[0], m[2]

    return dict(_f(o.strip()) for o in s.splitlines() if o.strip() and not re.match(r'\s*#', o))

# %% ../nbs/03_xtras.ipynb #d030f939
def expand_wildcards(code):
    "Expand all wildcard imports in the given code string."
    import ast,importlib
    tree = ast.parse(code)

    def _replace_node(code, old_node, new_node):
        "Replace `old_node` in the source `code` with `new_node`."
        lines = code.splitlines()
        lnum = old_node.lineno
        indent = ' ' * (len(lines[lnum-1]) - len(lines[lnum-1].lstrip()))
        new_lines = [indent+line for line in ast.unparse(new_node).splitlines()]
        lines[lnum-1 : old_node.end_lineno] = new_lines
        return '\n'.join(lines)

    def _expand_import(node, mod, existing):
        "Create expanded import `node` in `tree` from wildcard import of `mod`."
        mod_all = getattr(mod, '__all__', None)
        available_names = set(mod_all) if mod_all is not None else set(dir(mod))
        used_names = {n.id for n in ast.walk(tree) if isinstance(n, ast.Name) and n.id in available_names} - existing
        if not used_names: return node
        names = [ast.alias(name=name, asname=None) for name in sorted(used_names)]
        return ast.ImportFrom(module=node.module, names=names, level=node.level)

    existing = set()
    for node in ast.walk(tree):
        if isinstance(node, ast.ImportFrom) and node.names[0].name != '*': existing.update(n.name for n in node.names)
        elif isinstance(node, ast.Import): existing.update(n.name.split('.')[0] for n in node.names)
    for node in ast.walk(tree):
        if isinstance(node, ast.ImportFrom) and any(n.name == '*' for n in node.names):
            new_import = _expand_import(node, importlib.import_module(node.module), existing)
            code = _replace_node(code, node, new_import)
    return code

# %% ../nbs/03_xtras.ipynb #15c2b62c
@patch
def mkdir_perms(self:Path, mode=0o777, parents=False, exist_ok=False, uid=-1, gid=-1):
    "Create directory like Path.mkdir but optionally set uid/gid on newly created dirs"
    if self.exists() and not exist_ok: raise FileExistsError(self)
    to_create = []
    p = self
    while not p.exists():
        to_create.append(p)
        if parents: p = p.parent
        else: break
    for d in reversed(to_create):
        try: d.mkdir(mode=mode)
        except FileExistsError:
            if d==self and not exist_ok: raise
            continue
        if uid>-1 or gid>-1: os.chown(d, uid, gid)

# %% ../nbs/03_xtras.ipynb #832397a3
@contextmanager
def atomic_save(fn, mode='wb', uid=-1, gid=-1, **kwargs):
    "Context manager for writing a file atomically via a temp file that is renamed on close"
    from tempfile import NamedTemporaryFile as ntf
    fn = Path(fn)
    fn.parent.mkdir_perms(parents=True, exist_ok=True, uid=uid, gid=gid)
    with ntf(mode=mode, dir=fn.parent, delete=False, suffix=fn.suffix, **kwargs) as f:
        try: yield f
        except:
            Path(f.name).unlink(missing_ok=True)
            raise
    if uid>-1 or gid>-1: os.chown(f.name, uid, gid)
    Path(f.name).rename(fn)


# %% ../nbs/03_xtras.ipynb #9579358d
def dict2obj(d=None, list_func=L, dict_func=AttrDict, **kwargs):
    "Convert (possibly nested) dicts (or lists of dicts) to `AttrDict`"
    if d is None: d={}
    if isinstance(d, (L,list)): return list_func([dict2obj(v, list_func=list_func, dict_func=dict_func) for v in d])
    if not isinstance(d, dict): return d
    return dict_func(**{k:dict2obj(v, list_func=list_func, dict_func=dict_func) for k,v in (d|kwargs).items()})

# %% ../nbs/03_xtras.ipynb #055a1e07
def obj2dict(d):
    "Convert (possibly nested) AttrDicts (or lists of AttrDicts) to `dict`"
    if isinstance(d, (L,list)): return list(L(d).map(obj2dict))
    if not isinstance(d, dict): return d
    return dict(**{k:obj2dict(v) for k,v in d.items()})

# %% ../nbs/03_xtras.ipynb #36b00c72
def _repr_dict(d, lvl):
    if isinstance(d,dict):
        its = [f"{k}: {_repr_dict(v,lvl+1)}" for k,v in d.items()]
    elif isinstance(d,(list,L)): its = [_repr_dict(o,lvl+1) for o in d]
    else: return str(d)
    return '\n' + '\n'.join([" "*(lvl*2) + "- " + o for o in its])

# %% ../nbs/03_xtras.ipynb #13bb8fba
def repr_dict(d):
    "Print nested dicts and lists, such as returned by `dict2obj`"
    return _repr_dict(d,0).strip()

# %% ../nbs/03_xtras.ipynb #20842e1f
def is_listy(x):
    "`isinstance(x, (tuple,list,L,slice,Generator))`"
    return isinstance(x, (tuple,list,L,slice,Generator))

# %% ../nbs/03_xtras.ipynb #40c2c641
def mapped(f, it):
    "map `f` over `it`, unless it's not listy, in which case return `f(it)`"
    return L(it).map(f) if is_listy(it) else f(it)

# %% ../nbs/03_xtras.ipynb #ab0758e6
@patch
def readlines(self:Path, hint=-1, encoding='utf8'):
    "Read the content of `self`"
    with self.open(encoding=encoding) as f: return f.readlines(hint)

# %% ../nbs/03_xtras.ipynb #5ad0571d
@patch
def read_json(self:Path, encoding=None, errors=None):
    "Same as `read_text` followed by `loads`"
    return loads(self.read_text(encoding=encoding, errors=errors))

# %% ../nbs/03_xtras.ipynb #d6d8d893
@patch
def mk_write(self:Path, data, encoding=None, errors=None, mode=511, uid=-1, gid=-1):
    "Make all parent dirs of `self`, and write `data`"
    self.parent.mkdir(exist_ok=True, parents=True, mode=mode)
    self.write_text(data, encoding=encoding, errors=errors)
    if uid!=-1 or gid!=-1: os.chown(self, uid, gid)

# %% ../nbs/03_xtras.ipynb #233a8732
@patch
def write_json(self:Path, data, encoding=None, errors=None, mode=511, uid=-1, gid=-1, **kw):
    "Same as `dumps`followed by `mk_write`"
    self.mk_write(dumps(data,**kw),encoding,errors,mode,uid,gid)

# %% ../nbs/03_xtras.ipynb #9fc9965f
@patch
def relpath(self:Path, start=None):
    "Same as `os.path.relpath`, but returns a `Path`, and resolves symlinks"
    return Path(os.path.relpath(self.resolve(), Path(start).resolve()))

# %% ../nbs/03_xtras.ipynb #12264861
@patch
def ls(self:Path, n_max=None, file_type=None, file_exts=None):
    "Contents of path as a list"
    import mimetypes
    extns=L(file_exts)
    if file_type: extns += L(k for k,v in mimetypes.types_map.items() if v.startswith(file_type+'/'))
    has_extns = len(extns)==0
    res = (o for o in self.iterdir() if has_extns or o.suffix in extns)
    if n_max is not None: res = itertools.islice(res, n_max)
    return L(res)

# %% ../nbs/03_xtras.ipynb #abe052c6
@patch
def normpath(self:Path):
    "Normalize path, eliminating double slashes, etc."
    return Path(os.path.normpath(self))

# %% ../nbs/03_xtras.ipynb #695f90e0
@patch
def __repr__(self:Path):
    b = getattr(Path, 'BASE_PATH', None)
    if b:
        try: self = self.relative_to(b)
        except: pass
    return f"Path({self.as_posix()!r})"

# %% ../nbs/03_xtras.ipynb #6bd8b653
@patch
def delete(self:Path):
    "Delete a file, symlink, or directory tree"
    if not self.exists(): return
    if self.is_dir():
        import shutil
        shutil.rmtree(self)
    else: self.unlink()

# %% ../nbs/03_xtras.ipynb #f2bb033e
class IterLen:
    "Base class to add iteration to anything supporting `__len__` and `__getitem__`"
    def __iter__(self): return (self[i] for i in range_of(self))

# %% ../nbs/03_xtras.ipynb #0356640b
@docs
class ReindexCollection(GetAttr, IterLen):
    "Reindexes collection `coll` with indices `idxs` and optional LRU cache of size `cache`"
    _default='coll'
    def __init__(self, coll, idxs=None, cache=None, tfm=noop):
        if idxs is None: idxs = L.range(coll)
        store_attr()
        if cache is not None: self._get = functools.lru_cache(maxsize=cache)(self._get)

    def _get(self, i): return self.tfm(self.coll[i])
    def __getitem__(self, i): return self._get(self.idxs[i])
    def __len__(self): return len(self.coll)
    def reindex(self, idxs): self.idxs = idxs
    def shuffle(self):
        import random
        random.shuffle(self.idxs)
    def cache_clear(self): self._get.cache_clear()
    def __getstate__(self): return {'coll': self.coll, 'idxs': self.idxs, 'cache': self.cache, 'tfm': self.tfm}
    def __setstate__(self, s): self.coll,self.idxs,self.cache,self.tfm = s['coll'],s['idxs'],s['cache'],s['tfm']

    _docs = dict(reindex="Replace `self.idxs` with idxs",
                shuffle="Randomly shuffle indices",
                cache_clear="Clear LRU cache")

# %% ../nbs/03_xtras.ipynb #59a2f54d
class SaveReturn:
    "Wrap an iterator such that the generator function's return value is stored in `.value`"
    def __init__(self, its): self.its = its
    def __iter__(self):
        self.value = yield from self.its
        return self.value

# %% ../nbs/03_xtras.ipynb #e89ff9b0
def trim_wraps(f, n=1):
    "Like wraps, but removes the first n parameters from the signature"
    import inspect
    def _(g):
        g = wraps(f)(g)
        sig = inspect.signature(f)
        params = list(sig.parameters.values())[n:]
        g.__signature__ = sig.replace(parameters=params)
        return g
    return _

# %% ../nbs/03_xtras.ipynb #b867efdd
class _save_iter:
    def __init__(self, g, *args, **kw): self.g,self.args,self.kw = g,args,kw
    def __iter__(self): yield from self.g(self, *self.args, **self.kw)
    def __aiter__(self): return self.g(self, *self.args, **self.kw)

def save_iter(g):
    "Decorator that allows a generator function to store values in the returned iterator object"
    @trim_wraps(g)
    def _(*args, **kwargs): return _save_iter(g, *args, **kwargs)
    return _

# %% ../nbs/03_xtras.ipynb #d4017861
def asave_iter(g):
    "Like `save_iter`, but for async iterators"
    @trim_wraps(g)
    def _(*args, **kwargs): return _save_iter(g, *args, **kwargs)
    return _

# %% ../nbs/03_xtras.ipynb #20e906c7
def frontmatter(txt:str)->dict:
    "Dict contained in frontmatter in `txt`, if present"
    import yaml
    if not txt.startswith('---\n'): return {}
    fm,part,_ = txt[4:].partition('\n---\n')
    if not part: return {}
    try: res = yaml.safe_load(fm)
    except yaml.parser.ParserError: return {}
    return res if isinstance(res,dict) else {}

# %% ../nbs/03_xtras.ipynb #45eb5141
def clean_cli_output(txt:str, strip:bool=True):
    "Clean CLI output by handling alternate screen, carriage returns, and ANSI escapes"
    if '\x1b[?1049h' in txt:
        if '\x1b[?1049l' not in txt: return ''
        txt = txt.rsplit('\x1b[?1049l', 1)[-1]
    txt = txt.replace('\r\n', '\n')
    res = '\n'.join(l.rsplit('\r', 1)[-1] for l in txt.split('\n'))
    return strip_ansi(res, term_queries=True) if strip else res

# %% ../nbs/03_xtras.ipynb #35368de5
def unqid(seeded=False):
    "Generate a unique id suitable for use as a Python identifier"
    from base64 import b64encode
    from uuid import uuid4,UUID
    id4 = UUID(int=random.getrandbits(128), version=4) if seeded else uuid4()
    res = b64encode(id4.bytes)
    return '_' + res.decode().rstrip('=').translate(str.maketrans('+/', '_-'))

# %% ../nbs/03_xtras.ipynb #e0e7b294
def rtoken_hex(
    nbytes:int=16 # Number of bytes to generate
)->str: # hex string of length nbytes*2
    "Generate a random hex string using Python's random module."
    import random
    return ''.join(f'{random.randint(0, 255):02x}' for _ in range(nbytes))

# %% ../nbs/03_xtras.ipynb #4e02b1f0
def friendly_name(levels=3, suffix=4):
    "Generate a random human-readable name with customizable word levels and suffix length"
    import random,string
    adjectives = ['admiring', 'adoring', 'amazing', 'awesome', 'beautiful', 'blissful', 'bold', 'brave', 'busy', 'charming', 'clever', 'compassionate', 'confident', 'cool', 'dazzling', 'determined', 'dreamy', 'eager', 'ecstatic', 'elastic', 'elated', 'elegant', 'epic', 'exciting', 'fervent', 'festive', 'flamboyant', 'focused', 'friendly', 'frosty', 'funny', 'gallant', 'gifted', 'goofy', 'gracious', 'great', 'happy', 'hopeful', 'hungry', 'inspiring', 'intelligent', 'interesting', 'jolly', 'jovial', 'keen', 'kind', 'laughing', 'loving', 'lucid', 'magical', 'modest', 'nice', 'nifty', 'nostalgic', 'objective', 'optimistic', 'peaceful', 'pensive', 'practical', 'priceless', 'quirky', 'quizzical', 'relaxed', 'reverent', 'romantic', 'serene', 'sharp', 'silly', 'sleepy', 'stoic', 'sweet', 'tender', 'trusting', 'upbeat', 'vibrant', 'vigilant', 'vigorous', 'wizardly', 'wonderful', 'youthful', 'zealous', 'zen', 'golden', 'silver', 'crimson', 'azure', 'emerald', 'violet', 'amber', 'coral', 'turquoise', 'lavender', 'minty', 'citrus', 'vanilla', 'woody', 'floral', 'fresh', 'gentle', 'sparkling', 'precise', 'curious']
    nouns = ['tiger', 'eagle', 'river', 'mountain', 'forest', 'ocean', 'star', 'moon', 'wind', 'dragon', 'phoenix', 'wolf', 'bear', 'lion', 'shark', 'falcon', 'raven', 'crystal', 'diamond', 'ruby', 'sapphire', 'pearl', 'wave', 'tide', 'cloud', 'rainbow', 'sunset', 'sunrise', 'galaxy', 'comet', 'meteor', 'planet', 'nebula', 'cosmos', 'universe', 'atom', 'photon', 'quantum', 'matrix', 'cipher', 'code', 'signal', 'pulse', 'beam', 'ray', 'spark', 'frost', 'ice', 'snow', 'mist', 'fog', 'dew', 'rain', 'hail', 'helix', 'prism', 'lens', 'mirror', 'echo', 'heart', 'mind', 'dream', 'vision', 'hope', 'wish', 'magic', 'spell', 'charm', 'rune', 'symbol', 'token', 'key', 'door', 'gate', 'bridge', 'tower', 'castle', 'fortress', 'shield', 'dolphin', 'whale', 'penguin', 'butterfly', 'hummingbird', 'deer', 'rabbit', 'fox', 'otter', 'panda', 'koala', 'zebra', 'giraffe', 'elephant', 'valley', 'canyon', 'meadow', 'prairie', 'island', 'lake', 'pond', 'stream', 'waterfall', 'cliff', 'peak', 'hill', 'grove', 'garden', 'sunlight', 'breeze', 'melody', 'sparkle', 'whirlpool', 'windmill', 'carousel', 'spiral', 'glow']
    verbs = ['runs', 'flies', 'jumps', 'builds', 'creates', 'flows', 'shines', 'grows', 'moves', 'works', 'dances', 'sings', 'plays', 'dreams', 'thinks', 'learns', 'teaches', 'helps', 'heals', 'saves', 'protects', 'guards', 'watches', 'sees', 'hears', 'feels', 'knows', 'understands', 'discovers', 'explores', 'searches', 'finds', 'seeks', 'holds', 'carries', 'lifts', 'pushes', 'pulls', 'makes', 'crafts', 'forges', 'shapes', 'forms', 'molds', 'carves', 'joins', 'connects', 'links', 'binds', 'ties', 'opens', 'closes', 'starts', 'stops', 'begins', 'ends', 'finishes', 'completes', 'wins', 'triumphs', 'succeeds', 'achieves', 'accomplishes', 'reaches', 'arrives', 'departs', 'leaves', 'returns', 'comes', 'goes', 'travels', 'journeys', 'walks', 'sprints', 'races', 'speeds', 'rushes', 'hurries', 'waits', 'pauses', 'rests', 'sleeps', 'wakes', 'rises', 'climbs', 'ascends', 'descends', 'swims', 'dives', 'surfs', 'sails', 'paddles', 'hikes', 'treks', 'wanders', 'roams', 'ventures', 'navigates', 'glides', 'soars', 'floats', 'drifts', 'tosses', 'divides', 'shares', 'secures', 'settles', 'places', 'wonders', 'questions']
    adverbs = ['quickly', 'gracefully', 'gently', 'boldly', 'quietly', 'swiftly', 'carefully', 'eagerly', 'smoothly', 'brightly', 'softly', 'steadily', 'cleverly', 'proudly', 'calmly', 'freely', 'wisely', 'kindly', 'firmly', 'lightly', 'deeply', 'clearly', 'warmly', 'coolly', 'sharply', 'slowly', 'rapidly', 'silently', 'loudly', 'perfectly']
    suffix = ''.join(random.choices(string.ascii_lowercase + string.digits, k=suffix))
    words = [random.choice(o) for o in (adjectives,nouns,verbs,adverbs)[:levels]]
    return f"{'-'.join(words)}-{suffix}"

# %% ../nbs/03_xtras.ipynb #998e08a0
def n_friendly_names(levels=3, suffix=4):
    "Number of possible combos for `friendly_names"
    ns = [102,116,110,30]
    return product(ns[:levels])*36**suffix

# %% ../nbs/03_xtras.ipynb #b4d1831f
def exec_eval(code,   # Code to exec/eval
              g=None, # Globals namespace dict
              l=None  # Locals namespace dict
             ):
    "Evaluate `code` in `g` (defaults to `globals()`) and `l` (defaults to `locals()`)"
    import ast, inspect
    frame = inspect.currentframe().f_back
    if g is None: g = frame.f_globals
    if l is None: l = g
    tree = ast.parse(code, mode='exec')
    if tree.body and isinstance(tree.body[-1], ast.Expr):
        *statements, expr = tree.body
        exec_code = compile(ast.Module(statements, []), filename="<string>", mode="exec")
        expr_code = compile(ast.Expression(expr.value), filename="<string>", mode="eval")
        exec(exec_code, g, l)
        return eval(expr_code, g, l)
    else: exec(code, g, l)

# %% ../nbs/03_xtras.ipynb #bdaa78a0
def _unwrapped_type_dispatch_func(x): 
    # use isinstance_str to avoid adding plum-dispatch as dependency to fastcore
    return x.methods[0].implementation if isinstance_str(x,"Function") else x

# %% ../nbs/03_xtras.ipynb #bb4d3fe2
def _is_property(x): return type(x)==property
def _has_property_getter(x): return _is_property(x) and hasattr(x, 'fget') and hasattr(x.fget, 'func')
def _property_getter(x): return x.fget.func if _has_property_getter(x) else x

def _unwrapped_func(x):
    x = _unwrapped_type_dispatch_func(x)
    x = _property_getter(x)
    return x


def get_source_link(func):
    "Return link to `func` in source code"
    import inspect
    func = _unwrapped_func(func)
    try: line = inspect.getsourcelines(func)[1]
    except Exception: return ''
    mod = inspect.getmodule(func)
    module = mod.__name__.replace('.', '/') + '.py'
    try:
        nbdev_mod = import_module(mod.__package__.split('.')[0] + '._modidx')
        ghurl = nested_idx(nbdev_mod.d, 'settings', 'git_url')
        return f"{ghurl}{module}#L{line}"
    except: return f"{module}#L{line}"

# %% ../nbs/03_xtras.ipynb #31a380eb
spark_chars = '▁▂▃▅▆▇'

# %% ../nbs/03_xtras.ipynb #a6371687
def _ceil(x, lim=None): return x if (not lim or x <= lim) else lim

def _sparkchar(x, mn, mx, incr, empty_zero):
    if x is None or (empty_zero and not x): return ' '
    if incr == 0: return spark_chars[0]
    res = int((_ceil(x,mx)-mn)/incr-0.5)
    return spark_chars[res]

# %% ../nbs/03_xtras.ipynb #189818ca
def sparkline(data, mn=None, mx=None, empty_zero=False):
    "Sparkline for `data`, with `None`s (and zero, if `empty_zero`) shown as empty column"
    valid = [o for o in data if o is not None]
    if not valid: return ' '
    mn,mx,n = ifnone(mn,min(valid)),ifnone(mx,max(valid)),len(spark_chars)
    res = [_sparkchar(x=o, mn=mn, mx=mx, incr=(mx-mn)/n, empty_zero=empty_zero) for o in data]
    return ''.join(res)

# %% ../nbs/03_xtras.ipynb #81d193dd
def modify_exception(
    e:Exception, # An exception
    msg:str=None, # A custom message
    replace:bool=False, # Whether to replace e.args with [msg]
) -> Exception:
    "Modifies `e` with a custom message attached"
    e.args = [f'{e.args[0]} {msg}'] if not replace and len(e.args) > 0 else [msg]
    return e

# %% ../nbs/03_xtras.ipynb #6eed0273
def round_multiple(x, mult, round_down=False):
    "Round `x` to nearest multiple of `mult`"
    def _f(x_): return (int if round_down else round)(x_/mult)*mult
    res = L(x).map(_f)
    return res if is_listy(x) else res[0]

# %% ../nbs/03_xtras.ipynb #aa966f19
def set_num_threads(nt):
    "Get numpy (and others) to use `nt` threads"
    try: import mkl; mkl.set_num_threads(nt)
    except: pass
    try: import torch; torch.set_num_threads(nt)
    except: pass
    os.environ['IPC_ENABLE']='1'
    for o in ['OPENBLAS_NUM_THREADS','NUMEXPR_NUM_THREADS','OMP_NUM_THREADS','MKL_NUM_THREADS']:
        os.environ[o] = str(nt)

# %% ../nbs/03_xtras.ipynb #4dac8b41
def join_path_file(file, path, ext=''):
    "Return `path/file` if file is a string or a `Path`, file otherwise"
    if not isinstance(file, (str, Path)): return file
    path.mkdir(parents=True, exist_ok=True)
    return path/f'{file}{ext}'

# %% ../nbs/03_xtras.ipynb #205987ad
def autostart(g):
    "Decorator that automatically starts a generator"
    @functools.wraps(g)
    def f():
        r = g()
        next(r)
        return r
    return f

# %% ../nbs/03_xtras.ipynb #713d7560
class EventTimer:
    "An event timer with history of `store` items of time `span`"

    def __init__(self, store=5, span=60):
        import collections
        self.hist,self.span,self.last = collections.deque(maxlen=store),span,perf_counter()
        self._reset()

    def _reset(self): self.start,self.events = self.last,0

    def add(self, n=1):
        "Record `n` events"
        if self.duration>self.span:
            self.hist.append(self.freq)
            self._reset()
        self.events +=n
        self.last = perf_counter()

    @property
    def duration(self): return perf_counter()-self.start
    @property
    def freq(self): return self.events/self.duration

# %% ../nbs/03_xtras.ipynb #a79aa5d5
_fmt = string.Formatter()

# %% ../nbs/03_xtras.ipynb #b5bc6dd4
def stringfmt_names(s:str)->list:
    "Unique brace-delimited names in `s`"
    return uniqueify(o[1] for o in _fmt.parse(s) if o[1])

# %% ../nbs/03_xtras.ipynb #14ea0571
class PartialFormatter(string.Formatter):
    "A `string.Formatter` that doesn't error on missing fields, and tracks missing fields and unused args"
    def __init__(self):
        self.missing = set()
        super().__init__()

    def get_field(self, nm, args, kwargs):
        try: return super().get_field(nm, args, kwargs)
        except KeyError:
            self.missing.add(nm)
            return '{'+nm+'}',nm

    def check_unused_args(self, used, args, kwargs):
        self.xtra = filter_keys(kwargs, lambda o: o not in used)

# %% ../nbs/03_xtras.ipynb #eb9cb738
def partial_format(s:str, **kwargs):
    "string format `s`, ignoring missing field errors, returning missing and extra fields"
    fmt = PartialFormatter()
    res = fmt.format(s, **kwargs)
    return res,list(fmt.missing),fmt.xtra

# %% ../nbs/03_xtras.ipynb #becf7a2d
def truncstr(s:str, maxlen:int, suf:str='…', space='', sizevar:str=None)->str:
    "Truncate `s` to length `maxlen`, adding suffix `suf` if truncated"
    if sizevar: suf = suf.format_map({sizevar: len(s)})
    return s[:maxlen-len(suf)]+suf if len(s)+len(space)>maxlen else s+space

# %% ../nbs/03_xtras.ipynb #32fa9ba8
def utc2local(dt:datetime)->datetime:
    "Convert `dt` from UTC to local time"
    return dt.replace(tzinfo=timezone.utc).astimezone(tz=None)

# %% ../nbs/03_xtras.ipynb #55abbab7
def local2utc(dt:datetime)->datetime:
    "Convert `dt` from local to UTC time"
    return dt.replace(tzinfo=None).astimezone(tz=timezone.utc)

# %% ../nbs/03_xtras.ipynb #f32ab7a7
def trace(f):
    "Add `set_trace` to an existing function `f`"
    from pdb import set_trace
    if getattr(f, '_traced', False): return f
    def _inner(*args,**kwargs):
        set_trace()
        return f(*args,**kwargs)
    _inner._traced = True
    return _inner

# %% ../nbs/03_xtras.ipynb #45f779a3
@contextmanager
def modified_env(*delete, **replace):
    "Context manager temporarily modifying `os.environ` by deleting `delete` and replacing `replace`"
    prev = dict(os.environ)
    try:
        os.environ.update(replace)
        for k in delete: os.environ.pop(k, None)
        yield
    finally:
        os.environ.clear()
        os.environ.update(prev)

# %% ../nbs/03_xtras.ipynb #12f93aed
class ContextManagers(GetAttr):
    "Wrapper for `contextlib.ExitStack` which enters a collection of context managers"
    def __init__(self, mgrs): self.default,self.stack = L(mgrs),ExitStack()
    def __enter__(self): self.default.map(self.stack.enter_context)
    def __exit__(self, *args, **kwargs): self.stack.__exit__(*args, **kwargs)

# %% ../nbs/03_xtras.ipynb #f58a49b6
def shufflish(x, pct=0.04):
    "Randomly relocate items of `x` up to `pct` of `len(x)` from their starting location"
    n = len(x)
    import random
    return L(x[i] for i in sorted(range_of(x), key=lambda o: o+n*(1+random.random()*pct)))

# %% ../nbs/03_xtras.ipynb #3a527c27
def console_help(libname: str):
    "Show help for all console scripts from `libname`"
    from fastcore.style import S
    from importlib.metadata import entry_points
    for e in entry_points(group='console_scripts'):
        if e.module == libname or e.module.startswith(libname + '.'):
            nm = S.bold.light_blue(e.name)
            print(f'{nm:45}{e.load().__doc__}')

# %% ../nbs/03_xtras.ipynb #01a02b78
def hl_md(s, lang='html', show=True):
    "Syntax highlight `s` using `lang`."
    if hasattr(s, '__html__'): s = s.__html__()
    md = f'```{lang}\n{s}\n```'
    if not show: return md
    try:
        from IPython import display
        return display.Markdown(md)
    except ImportError: print(s)

# %% ../nbs/03_xtras.ipynb #b59609d4
def type2str(typ:type)->str:
    "Stringify `typ`"
    if typ is None or typ is NoneType: return 'None'
    if hasattr(typ, '__origin__'):
        args = ", ".join(type2str(arg) for arg in typ.__args__)
        if typ.__origin__ is Union: return f"Union[{args}]"
        return f"{typ.__origin__.__name__}[{args}]"
    elif isinstance(typ, type): return typ.__name__
    return str(typ)

# %% ../nbs/03_xtras.ipynb #9ef13805
def dataclass_src(cls):
    src = f"@dataclass\nclass {cls.__name__}:\n"
    for f in dataclasses.fields(cls):
        d = "" if f.default is dataclasses.MISSING else f" = {f.default!r}"
        src += f"    {f.name}: {type2str(f.type)}{d}\n"
    return src

# %% ../nbs/03_xtras.ipynb #ecd7f0a2
class Unset(Enum):
    _Unset=''
    def __repr__(self): return 'UNSET'
    def __str__ (self): return 'UNSET'
    def __bool__(self): return False
UNSET = Unset._Unset

# %% ../nbs/03_xtras.ipynb #1e9a46ab
def nullable_dc(cls):
    "Like `dataclass`, but default of `UNSET` added to fields without defaults"
    for k,v in get_annotations_ex(cls)[0].items():
        if not hasattr(cls,k): setattr(cls, k, field(default=UNSET))
    return dataclass(cls)

# %% ../nbs/03_xtras.ipynb #e62e3c77
def make_nullable(clas):
    if hasattr(clas, '_nullable'): return
    clas._nullable = True

    original_init = clas.__init__
    def __init__(self, *args, **kwargs):
        flds = fields(clas)
        dargs = {k.name:v for k,v in zip(flds, args)}
        for f in flds:
            nm = f.name
            if nm not in dargs and nm not in kwargs and f.default is None and f.default_factory is MISSING:
                kwargs[nm] = UNSET
        original_init(self, *args, **kwargs)
    
    clas.__init__ = __init__

    for f in fields(clas):
        if f.default is MISSING and f.default_factory is MISSING: f.default = None
    
    return clas

# %% ../nbs/03_xtras.ipynb #9fa70a25
def flexiclass(
        cls # The class to convert
    ) -> dataclass:
    "Convert `cls` into a `dataclass` like `make_nullable`. Converts in place and also returns the result."
    if is_dataclass(cls): return make_nullable(cls)
    for k,v in get_annotations_ex(cls)[0].items():
        if not hasattr(cls,k) or getattr(cls,k) is MISSING:
            setattr(cls, k, field(default=UNSET))
    return dataclass(cls, init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False)

# %% ../nbs/03_xtras.ipynb #7729d860
def asdict(o)->dict:
    "Convert `o` to a `dict`, supporting dataclasses, namedtuples, iterables, and `__dict__` attrs."
    if isinstance(o, dict): return o
    if hasattr(o, '_asdict'): r = o._asdict()
    elif is_dataclass(o): r = dataclasses.asdict(o)
    elif hasattr(o, '__iter__'):
        try: r = dict(o)
        except TypeError: pass
    elif hasattr(o, '__flds__'): r = {x:getattr(o,x) for x in o.__flds__}
    else: r = vars(o)
    skip = set(getattr(o, '__skip__', []))
    return {k:v for k,v in r.items() if v is not UNSET and v is not MISSING and k not in skip}

# %% ../nbs/03_xtras.ipynb #674a899f
def vars_pub(x):
    "Get public non-skipped vars"
    skip = set(getattr(x, '__skip__', []))
    return [o for o in vars(x) if o[0]!='_' and o not in skip]

# %% ../nbs/03_xtras.ipynb #575add44
def is_typeddict(cls:type)->bool:
    "Check if `cls` is a `TypedDict`"
    attrs = 'annotations', 'required_keys', 'optional_keys'
    return isinstance(cls, type) and all(hasattr(cls, f'__{attr}__') for attr in attrs)

# %% ../nbs/03_xtras.ipynb #557c2708
def is_namedtuple(cls):
    "`True` if `cls` is a namedtuple type"
    return issubclass(cls, tuple) and hasattr(cls, '_fields')

# %% ../nbs/03_xtras.ipynb #bb02cce3
class CachedIter:
    "Cache the result returned by an iterator"
    def __init__(self, o): self.o,self.value = o,UNSET
    def __iter__(self):
        if self.value is UNSET: self.value = yield from self.o
        return self.value

# %% ../nbs/03_xtras.ipynb #64471925
class CachedAwaitable:
    "Cache the result from an awaitable"
    def __init__(self, o): self.o,self.value = o,UNSET
    def __await__(self):
        if self.value is UNSET: self.value = yield from self.o.__await__()
        return self.value

# %% ../nbs/03_xtras.ipynb #86cb7a86
def reawaitable(func:callable):
    "Wraps the result of an asynchronous function into an object which can be awaited more than once"
    @wraps(func)
    def _f(*args, **kwargs): return CachedAwaitable(func(*args, **kwargs))
    return _f

# %% ../nbs/03_xtras.ipynb #1fc78492
def is_async_callable(obj):
    "Check if `obj` is an async callable, handling `partial` wrappers and callable instances"
    # Implementation from Starlette; Copyright © 2018, Encode OSS Ltd.
    from asyncio import iscoroutinefunction
    while isinstance(obj, partial): obj = obj.func
    return iscoroutinefunction(obj) or (callable(obj) and iscoroutinefunction(obj.__call__))

# %% ../nbs/03_xtras.ipynb #f7a2e9ed
async def maybe_await(o):
    "Await `o` if needed, and return it"
    from inspect import isawaitable
    return await o if isawaitable(o) else o

# %% ../nbs/03_xtras.ipynb #02f9f070
async def noopa(x=None, *args, **kwargs):
    "Do nothing (async)"
    return x

# %% ../nbs/03_xtras.ipynb #d2b4fe09
def flexicache(*funcs, maxsize=128):
    "Like `lru_cache`, but customisable with policy `funcs`"
    import asyncio
    from collections import OrderedDict
    def _f(func):
        cache,states = OrderedDict(), [None]*len(funcs)
        def _cache_logic(key, execute_func):
            if key in cache:
                result,states = cache[key]
                if not any(f(state) for f,state in zip(funcs, states)):
                    cache[key] = cache.pop(key)
                    return result
                del cache[key]
            try: newres = execute_func()
            except:
                if key not in cache: raise
                cache[key] = cache.pop(key)
                return result
            cache[key] = (newres, [f(None) for f in funcs])
            if len(cache) > maxsize: cache.popitem(last=False)
            return newres

        @wraps(func)
        def wrapper(*args, **kwargs):
            return _cache_logic(f"{args} // {kwargs}", lambda: func(*args, **kwargs))

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            return await _cache_logic(f"{args} // {kwargs}", lambda: asyncio.ensure_future(func(*args, **kwargs)))

        return async_wrapper if asyncio.iscoroutinefunction(func) else wrapper
    return _f

# %% ../nbs/03_xtras.ipynb #cf131274
def time_policy(seconds):
    "A `flexicache` policy that expires cached items after `seconds` have passed"
    def policy(last_time):
        now = time()
        return now if last_time is None or now-last_time>seconds else None
    return policy

# %% ../nbs/03_xtras.ipynb #8387a8bf
def mtime_policy(filepath):
    "A `flexicache` policy that expires cached items after `filepath` modified-time changes"
    def policy(mtime):
        current_mtime = getmtime(filepath)
        return current_mtime if mtime is None or current_mtime>mtime else None
    return policy

# %% ../nbs/03_xtras.ipynb #1094d811
def timed_cache(seconds=60, maxsize=128):
    "Like `lru_cache`, but also with time-based eviction"
    return flexicache(time_policy(seconds), maxsize=maxsize)



"Filters for processing ANSI colors."

# Copyright (c) IPython Development Team.
# Modifications by Jeremy Howard.

import re
from html import escape

__all__ = ["strip_ansi", "ansi2html", "ansi2latex"]

_ANSI_RE = re.compile("\x1b\\[(.*?)([@-~])")
_ANSI_COLORS = ( "ansi-black", "ansi-red", "ansi-green", "ansi-yellow", "ansi-blue", "ansi-magenta", "ansi-cyan", "ansi-white", "ansi-black-intense",
    "ansi-red-intense", "ansi-green-intense", "ansi-yellow-intense", "ansi-blue-intense", "ansi-magenta-intense", "ansi-cyan-intense", "ansi-white-intense")

def strip_terminal_queries(text):
    # Remove OSC sequences (like background color queries)
    text = re.sub('\x1b\\][^\x07]*\x07', '', text)
    # Remove DSR sequences (device status reports)
    return re.sub(r'\x1b\[[0-9]*n', '', text)


def strip_ansi(source, term_queries:bool=False):
    "Remove ANSI escape codes from text."
    if term_queries: source = strip_terminal_queries(source)
    return _ANSI_RE.sub("", source)


def ansi2html(text):
    "Convert ANSI colors to HTML colors."
    text = escape(strip_terminal_queries(text))
    return _ansi2anything(text, _htmlconverter)


def ansi2latex(text):
    "Convert ANSI colors to LaTeX colors."
    return _ansi2anything(text, _latexconverter)


def _htmlconverter(fg, bg, bold, underline, inverse):
    "Return start and end tags for given foreground/background/bold/underline."
    if (fg, bg, bold, underline, inverse) == (None, None, False, False, False): return "", ""

    classes,styles = [],[]
    if inverse: fg, bg = bg, fg
    if isinstance(fg, int): classes.append(_ANSI_COLORS[fg] + "-fg")
    elif fg: styles.append("color: rgb({},{},{})".format(*fg))
    elif inverse: classes.append("ansi-default-inverse-fg")

    if isinstance(bg, int): classes.append(_ANSI_COLORS[bg] + "-bg")
    elif bg: styles.append("background-color: rgb({},{},{})".format(*bg))
    elif inverse: classes.append("ansi-default-inverse-bg")

    if bold: classes.append("ansi-bold")
    if underline: classes.append("ansi-underline")

    starttag = "<span"
    if classes: starttag += ' class="' + " ".join(classes) + '"'
    if styles: starttag += ' style="' + "; ".join(styles) + '"'
    starttag += ">"
    return starttag, "</span>"


def _latexconverter(fg, bg, bold, underline, inverse):
    "Return start and end markup given foreground/background/bold/underline."
    if (fg, bg, bold, underline, inverse) == (None, None, False, False, False): return "", ""
    starttag, endtag = "", ""
    if inverse: fg, bg = bg, fg
    if isinstance(fg, int):
        starttag += r"\textcolor{" + _ANSI_COLORS[fg] + "}{"
        endtag = "}" + endtag
    elif fg:
        # See http://tex.stackexchange.com/a/291102/13684
        starttag += r"\def\tcRGB{\textcolor[RGB]}\expandafter"
        starttag += r"\tcRGB\expandafter{{\detokenize{{{},{},{}}}}}{{".format(*fg)
        endtag = "}" + endtag
    elif inverse:
        starttag += r"\textcolor{ansi-default-inverse-fg}{"
        endtag = "}" + endtag

    if isinstance(bg, int):
        starttag += r"\setlength{\fboxsep}{0pt}"
        starttag += r"\colorbox{" + _ANSI_COLORS[bg] + "}{"
        endtag = r"\strut}" + endtag
    elif bg:
        starttag += r"\setlength{\fboxsep}{0pt}"
        # See http://tex.stackexchange.com/a/291102/13684
        starttag += r"\def\cbRGB{\colorbox[RGB]}\expandafter"
        starttag += r"\cbRGB\expandafter{{\detokenize{{{},{},{}}}}}{{".format(*bg)
        endtag = r"\strut}" + endtag
    elif inverse:
        starttag += r"\setlength{\fboxsep}{0pt}"
        starttag += r"\colorbox{ansi-default-inverse-bg}{"
        endtag = r"\strut}" + endtag

    if bold:
        starttag += r"\textbf{"
        endtag = "}" + endtag

    if underline:
        starttag += r"\underline{"
        endtag = "}" + endtag

    return starttag, endtag


def _ansi2anything(text, converter):
    "Convert ANSI colors to HTML or LaTeX."
    fg, bg = None, None
    bold, underline, inverse = False, False, False
    numbers,out = [],[]

    while text:
        m = _ANSI_RE.search(text)
        if m:
            if m.group(2) == "m":
                # Empty code is same as code 0
                try: numbers = [int(n) if n else 0 for n in m.group(1).split(";")]
                except ValueError: pass  # Invalid color specification
            else: pass  # Not a color code
            chunk, text = text[: m.start()], text[m.end() :]
        else: chunk, text = text, ""

        if chunk:
            starttag, endtag = converter(
                fg + 8 if bold and fg in range(8) else fg,  # type:ignore[operator]
                bg, bold, underline, inverse)
            out.append(starttag)
            out.append(chunk)
            out.append(endtag)

        while numbers:
            n = numbers.pop(0)
            if n == 0:
                # Code 0 (same as empty code): reset everything
                fg = bg = None
                bold = underline = inverse = False
            elif n == 1: bold = True
            elif n == 4: underline = True
            # Code 5: blinking
            elif n == 5: bold = True
            elif n == 7: inverse = True
            elif n in (21, 22): bold = False
            elif n == 24: underline = False
            elif n == 27: inverse = False
            elif 30 <= n <= 37: fg = n - 30
            elif n == 38:
                try: fg = _get_extended_color(numbers)
                except ValueError: numbers.clear()
            elif n == 39: fg = None
            elif 40 <= n <= 47: bg = n - 40
            elif n == 48:
                try: bg = _get_extended_color(numbers)
                except ValueError: numbers.clear()
            elif n == 49: bg = None
            elif 90 <= n <= 97: fg = n - 90 + 8
            elif 100 <= n <= 107: bg = n - 100 + 8
            else: pass  # Unknown codes are ignored
    return "".join(out)


def _get_extended_color(numbers):
    n = numbers.pop(0)
    if n == 2 and len(numbers) >= 3:
        # 24-bit RGB
        r = numbers.pop(0)
        g = numbers.pop(0)
        b = numbers.pop(0)
        if not all(0 <= c <= 255 for c in (r, g, b)): raise ValueError()
    elif n == 5 and len(numbers) >= 1:
        # 256 colors
        idx = numbers.pop(0)
        if idx < 0: raise ValueError()
        # 16 default terminal colors
        if idx < 16: return idx
        if idx < 232:
            # 6x6x6 color cube, see http://stackoverflow.com/a/27165165/500098
            r = (idx - 16) // 36
            r = 55 + r * 40 if r > 0 else 0
            g = ((idx - 16) % 36) // 6
            g = 55 + g * 40 if g > 0 else 0
            b = (idx - 16) % 6
            b = 55 + b * 40 if b > 0 else 0
        # grayscale, see http://stackoverflow.com/a/27165165/500098
        elif idx < 256: r = g = b = (idx - 232) * 10 + 8
        else: raise ValueError()
    else: raise ValueError()
    return r, g, b




"""The `L` class and helpers for it"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_foundation.ipynb.

# %% auto #0
__all__ = ['working_directory', 'add_docs', 'docs', 'coll_repr', 'is_bool', 'mask2idxs', 'cycle', 'zip_cycle', 'is_indexer',
           'product', 'flatmap', 'CollBase', 'L', 'curryable', 'splitter', 'linesplitter', 'save_config_file',
           'read_config_file', 'find_file_parents', 'Config']

# %% ../nbs/02_foundation.ipynb #0f974791
from .imports import *
from .basics import *
from functools import lru_cache,wraps
from contextlib import contextmanager
from copy import copy
from configparser import ConfigParser
import random,pickle,inspect

# %% ../nbs/02_foundation.ipynb #f059df0f
@contextmanager
def working_directory(path):
    "Change working directory to `path` and return to previous on exit."
    prev_cwd = Path.cwd()
    os.chdir(path)
    try: yield
    finally: os.chdir(prev_cwd)

# %% ../nbs/02_foundation.ipynb #c3fd027f
def add_docs(cls, cls_doc=None, **docs):
    "Copy values from `docs` to `cls` docstrings, and confirm all public methods are documented"
    if cls_doc is not None: cls.__doc__ = cls_doc
    for k,v in docs.items():
        f = getattr(cls,k)
        if hasattr(f,'__func__'): f = f.__func__ # required for class methods
        f.__doc__ = v
    # List of public callables without docstring
    nodoc = [c for n,c in vars(cls).items() if callable(c)
             and not n.startswith('_') and c.__doc__ is None]
    assert not nodoc, f"Missing docs: {nodoc}"
    assert cls.__doc__ is not None, f"Missing class docs: {cls}"

# %% ../nbs/02_foundation.ipynb #b180313a
def docs(cls):
    "Decorator version of `add_docs`, using `_docs` dict"
    add_docs(cls, **cls._docs)
    return cls

# %% ../nbs/02_foundation.ipynb #34a29b43
def coll_repr(c, max_n=250):
    "String repr of up to `max_n` items of (possibly lazy) collection `c`"
    res = ', '.join(itertools.islice(map(repr,c), max_n))
    res = '[' + res + ('...' if len(c)>max_n else '') + ']'
    if len(c)>=max(7,max_n): res = f'(#{len(c)}) ' + res
    return res


# %% ../nbs/02_foundation.ipynb #6cf96f49
def is_bool(x):
    "Check whether `x` is a bool or None"
    return isinstance(x,(bool,NoneType)) or risinstance('bool_', x)

# %% ../nbs/02_foundation.ipynb #66a28a53
def mask2idxs(mask):
    "Convert bool mask or index list to index `L`"
    if isinstance(mask,slice): return mask
    mask = list(mask)
    if len(mask)==0: return []
    it = mask[0]
    if hasattr(it,'item'): it = it.item()
    if is_bool(it): return [i for i,m in enumerate(mask) if m]
    return [int(i) for i in mask]

# %% ../nbs/02_foundation.ipynb #af754019
def cycle(o):
    "Like `itertools.cycle` except creates list of `None`s if `o` is empty"
    o = listify(o)
    return itertools.cycle(o) if o is not None and len(o) > 0 else itertools.cycle([None])

# %% ../nbs/02_foundation.ipynb #9346b91a
def zip_cycle(x, *args):
    "Like `itertools.zip_longest` but `cycle`s through elements of all but first argument"
    return zip(x, *map(cycle,args))

# %% ../nbs/02_foundation.ipynb #8a281173
def is_indexer(idx):
    "Test whether `idx` will index a single item in a list"
    return isinstance(idx,int) or not getattr(idx,'ndim',1)

# %% ../nbs/02_foundation.ipynb #3fd3133f
def product(xs):
    "The product of elements of `xs`, with `None`s removed"
    return reduce(operator.mul, [o for o in xs if o is not None], 1)

# %% ../nbs/02_foundation.ipynb #b1e1639e
def flatmap(f, xs, **kwargs):
    "Apply f to each element and flatten the results into a single list."
    return [y for x in xs for y in f(x, **kwargs)]

# %% ../nbs/02_foundation.ipynb #eefd2763
class CollBase:
    "Base class for composing a list of `items`"
    def __init__(self, items): self.items = items
    def __len__(self): return len(self.items)
    def __getitem__(self, k): return self.items[list(k) if isinstance(k,CollBase) else k]
    def __setitem__(self, k, v): self.items[list(k) if isinstance(k,CollBase) else k] = v
    def __delitem__(self, i): del(self.items[i])
    def __repr__(self): return self.items.__repr__()
    def __iter__(self): return self.items.__iter__()

# %% ../nbs/02_foundation.ipynb #d953eaf9
class _L_Meta(type):
    def __call__(cls, x=None, *args, **kwargs):
        if not args and not kwargs and x is not None and isinstance(x,cls): return x
        return super().__call__(x, *args, **kwargs)

# %% ../nbs/02_foundation.ipynb #f571f489
class L(GetAttr, CollBase, metaclass=_L_Meta):
    "Behaves like a list of `items` but can also index with list of indices or masks"
    _default='items'
    def __init__(self, items=None, *rest, use_list=False, match=None):
        if (use_list is not None) or not is_array(items):
            items = listify(items, *rest, use_list=use_list, match=match)
        super().__init__(items)

    @property
    def _xtra(self): return None
    def _new(self, items, *args, **kwargs): return type(self)(items, *args, use_list=None, **kwargs)
    def __getitem__(self, idx):
        "Retrieve `idx` (can be list of indices, or mask, or int) items"
        if isinstance(idx,int) and not hasattr(self.items,'iloc'): return self.items[idx]
        return self._get(idx) if is_indexer(idx) else L(self._get(idx), use_list=None)

    def _get(self, i):
        if is_indexer(i) or isinstance(i,slice): return getattr(self.items,'iloc',self.items)[i]
        i = mask2idxs(i)
        return (self.items.iloc[list(i)] if hasattr(self.items,'iloc')
                else self.items.__array__()[(i,)] if hasattr(self.items,'__array__')
                else [self.items[i_] for i_ in i])

    def __setitem__(self, idx, o):
        "Set `idx` (can be list of indices, or mask, or int) items to `o` (which is broadcast if not iterable)"
        if isinstance(idx, (int, slice)): self.items[idx] = o
        else:
            idx = idx if isinstance(idx,L) else listify(idx)
            if not is_iter(o): o = [o]*len(idx)
            for i,o_ in zip(idx,o): self.items[i] = o_

    def __eq__(self,b):
        if b is None: return False
        if not hasattr(b, '__iter__'): return False
        if risinstance('ndarray', b): return array_equal(b, self)
        if isinstance(b, (str,dict)) or callable(b): return False
        return all_equal(b,self)

    def __iter__(self): return iter(self.items.itertuples() if hasattr(self.items,'iloc') else self.items)
    def __contains__(self,b): return b in self.items
    def __reversed__(self): return self._new(reversed(self.items))
    def __invert__(self): return self._new(not i for i in self)
    def __repr__(self): return repr(self.items)
    def _repr_pretty_(self, p, cycle):
        p.text('...' if cycle else repr(self.items) if is_array(self.items) else coll_repr(self))
    def __mul__ (a,b): return a._new(a.items*b)
    def __add__ (a,b): return a._new(a.items+listify(b))
    def __radd__(a,b): return a._new(b)+a
    def __addi__(a,b):
        a.items += list(b)
        return a

# %% ../nbs/02_foundation.ipynb #05e35e65
# Here we are fixing the signature of L. What happens is that the __call__ method on the MetaClass of L shadows the __init__
# giving the wrong signature (https://stackoverflow.com/questions/49740290/call-from-metaclass-shadows-signature-of-init).
def _f(items=None, *rest, use_list=False, match=None): ...
L.__signature__ = inspect.signature(_f)

# %% ../nbs/02_foundation.ipynb #68e52115
Sequence.register(L);

# %% ../nbs/02_foundation.ipynb #d966b525
@patch
def unique(self:L, sort=False, bidir=False, start=None):
    "Unique items, in stable order"
    return L(uniqueify(self, sort=sort, bidir=bidir, start=start))

# %% ../nbs/02_foundation.ipynb #840aef37
@patch
def val2idx(self:L):
    "Dict from value to index"
    return val2idx(self)

# %% ../nbs/02_foundation.ipynb #f2e038cc
@patch(cls_method=True)
def range(cls:L, a, b=None, step=None):
    "Class Method: Same as `range`, but returns `L`. Can pass collection for `a`, to use `len(a)`"
    return cls(range_of(a, b=b, step=step))

# %% ../nbs/02_foundation.ipynb #d1d16238
@patch
def enumerate(self:L):
    "Same as `enumerate`"
    return L(enumerate(self))

# %% ../nbs/02_foundation.ipynb #ccd5c96c
@patch
def renumerate(self:L):
    "Same as `renumerate`"
    return L(renumerate(self))

# %% ../nbs/02_foundation.ipynb #13795373
@patch(cls_method=True)
def split(cls:L, s, sep=None, maxsplit=-1):
    "Class Method: Same as `str.split`, but returns an `L`"
    return cls(s.split(sep,maxsplit))

# %% ../nbs/02_foundation.ipynb #6da7b43c
@patch(cls_method=True)
def splitlines(cls:L, s, keepends=False):
    "Class Method: Same as `str.splitlines`, but returns an `L`"
    return cls(s.splitlines(keepends))

# %% ../nbs/02_foundation.ipynb #826ffa88
def curryable(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        if not isinstance(self, L): return lambda items: f(L(items), self, *args, **kwargs)
        return f(L(self), *args, **kwargs)
    return wrapper

# %% ../nbs/02_foundation.ipynb #0d6407c6
@patch
@curryable
def map(self:L, f, *args, **kwargs):
    "Create new `L` with `f` applied to all `items`, passing `args` and `kwargs` to `f`"
    return self._new(map_ex(self, f, *args, gen=False, **kwargs))

# %% ../nbs/02_foundation.ipynb #9c5a4633
def splitter(sep=None, maxsplit=-1):
    "Create a partial function that splits strings into `L`"
    return partial(L.split, sep=sep, maxsplit=maxsplit)

# %% ../nbs/02_foundation.ipynb #ce5b413b
def linesplitter(keepends=False):
    "Create a partial function that splits strings by lines into `L`"
    return partial(L.splitlines, keepends=keepends)

# %% ../nbs/02_foundation.ipynb #d2680cbd
@patch
@curryable
def groupby(self:L, key, val=noop):
    "Same as `fastcore.basics.groupby`"
    return groupby(self, key, val=val)

# %% ../nbs/02_foundation.ipynb #37dc972a
@patch
@curryable
def starmap(self:L, f, *args, **kwargs):
    "Like `map`, but use `itertools.starmap`"
    return self._new(itertools.starmap(partial(f,*args,**kwargs), self))

# %% ../nbs/02_foundation.ipynb #cb2283b0
@patch
@curryable
def rstarmap(self:L, f, *args, **kwargs):
    "Like `starmap`, but reverse the order of args"
    return self._new(itertools.starmap(lambda *x: f(*x[::-1], *args, **kwargs), self))

# %% ../nbs/02_foundation.ipynb #c0846f2e
@patch
def map_dict(self:L, f=noop, *args, **kwargs):
    "Like `map`, but creates a dict from `items` to function results"
    return {k:f(k, *args,**kwargs) for k in self}

# %% ../nbs/02_foundation.ipynb #105ac114
@patch
def zip(self:L, cycled=False):
    "Create new `L` with `zip(*items)`"
    return self._new((zip_cycle if cycled else zip)(*self))

# %% ../nbs/02_foundation.ipynb #6340b2b3
@patch
def map_zip(self:L, f, *args, cycled=False, **kwargs):
    "Combine `zip` and `starmap`"
    return self.zip(cycled=cycled).starmap(f, *args, **kwargs)

# %% ../nbs/02_foundation.ipynb #bc012fad
@patch
def zipwith(self:L, *rest, cycled=False):
    "Create new `L` with `self` zip with each of `*rest`"
    return self._new([self, *rest]).zip(cycled=cycled)

# %% ../nbs/02_foundation.ipynb #5a40226f
@patch
def map_zipwith(self:L, f, *rest, cycled=False, **kwargs):
    "Combine `zipwith` and `starmap`"
    return self.zipwith(*rest, cycled=cycled).starmap(f, **kwargs)

# %% ../nbs/02_foundation.ipynb #66e76d03
@patch
@curryable
def filter(self:L, f=noop, negate=False, **kwargs):
    "Create new `L` filtered by predicate `f`, passing `args` and `kwargs` to `f`"
    return self._new(filter_ex(self, f=f, negate=negate, gen=False, **kwargs))

# %% ../nbs/02_foundation.ipynb #085b690b
@patch
@curryable
def starfilter(self:L, f, negate=False, **kwargs):
    "Like `filter`, but unpacks elements as args to `f`"
    _f = lambda x: f(*x, **kwargs)
    if negate: _f = not_(_f)
    return self._new(filter(_f, self))

# %% ../nbs/02_foundation.ipynb #e0fa80df
@patch
@curryable
def rstarfilter(self:L, f, negate=False, **kwargs):
    "Like `starfilter`, but reverse the order of args"
    _f = lambda x: f(*x[::-1], **kwargs)
    if negate: _f = not_(_f)
    return self._new(filter(_f, self))

# %% ../nbs/02_foundation.ipynb #a2515ab8
@patch
@curryable
def argwhere(self:L, f, negate=False, **kwargs):
    "Like `filter`, but return indices for matching items"
    return self._new(argwhere(self, f, negate, **kwargs))

# %% ../nbs/02_foundation.ipynb #cf22acd4
@patch
@curryable
def starargwhere(self:L, f, negate=False):
    "Like `argwhere`, but unpacks elements as args to `f`"
    _f = lambda x: f(*x)
    if negate: _f = not_(_f)
    return self._new(i for i,o in enumerate(self) if _f(o))

# %% ../nbs/02_foundation.ipynb #1f010f9c
@patch
@curryable
def rstarargwhere(self:L, f, negate=False):
    "Like `starargwhere`, but reverse the order of args"
    _f = lambda x: f(*x[::-1])
    if negate: _f = not_(_f)
    return self._new(i for i,o in enumerate(self) if _f(o))

# %% ../nbs/02_foundation.ipynb #3e7fa556
@patch
@curryable
def argfirst(self:L, f, negate=False):
    "Return index of first matching item"
    if negate: f = not_(f)
    return first(i for i,o in self.enumerate() if f(o))

# %% ../nbs/02_foundation.ipynb #90509c05
@patch
@curryable
def starargfirst(self:L, f, negate=False):
    "Like `argfirst`, but unpacks elements as args to `f`"
    _f = lambda x: f(*x)
    if negate: _f = not_(_f)
    return first(i for i,o in self.enumerate() if _f(o))

# %% ../nbs/02_foundation.ipynb #dc43b9b9
@patch
@curryable
def rstarargfirst(self:L, f, negate=False):
    "Like `starargfirst`, but reverse the order of args"
    _f = lambda x: f(*x[::-1])
    if negate: _f = not_(_f)
    return first(i for i,o in self.enumerate() if _f(o))

# %% ../nbs/02_foundation.ipynb #44055cd0
@patch
def itemgot(self:L, *idxs):
    "Create new `L` with item `idx` of all `items`"
    x = self
    for idx in idxs: x = x.map(itemgetter(idx))
    return x

# %% ../nbs/02_foundation.ipynb #10f3d35e
@patch
def attrgot(self:L, k, default=None):
    "Create new `L` with attr `k` (or value `k` for dicts) of all `items`."
    return self.map(lambda o: o.get(k,default) if isinstance(o, dict) else nested_attr(o,k,default))

# %% ../nbs/02_foundation.ipynb #9f238dfb
@patch
@curryable
def sorted(self:L, key=None, reverse=False, cmp=None, **kwargs):
    "New `L` sorted by `key`, using `sort_ex`. If key is str use `attrgetter`; if int use `itemgetter`"
    return self._new(sorted_ex(self, key=key, reverse=reverse, cmp=cmp, **kwargs))

# %% ../nbs/02_foundation.ipynb #1c21919b
@patch
@curryable
def starsorted(self:L, key, reverse=False):
    "Like `sorted`, but unpacks elements as args to `key`"
    return self._new(sorted(self, key=lambda x: key(*x), reverse=reverse))

# %% ../nbs/02_foundation.ipynb #7549144e
@patch
@curryable
def rstarsorted(self:L, key, reverse=False):
    "Like `starsorted`, but reverse the order of args"
    return self._new(sorted(self, key=lambda x: key(*x[::-1]), reverse=reverse))

# %% ../nbs/02_foundation.ipynb #9d86b896
@patch
def concat(self:L):
    "Concatenate all elements of list"
    return self._new(itertools.chain.from_iterable(self.map(L)))

# %% ../nbs/02_foundation.ipynb #b493aad4
@patch
def copy(self:L):
    "Same as `list.copy`, but returns an `L`"
    return self._new(self.items.copy())

# %% ../nbs/02_foundation.ipynb #7b2691d6
@patch
def shuffle(self:L):
    "Same as `random.shuffle`, but not inplace"
    it = copy(self.items)
    random.shuffle(it)
    return self._new(it)

# %% ../nbs/02_foundation.ipynb #7b83c9e9
@patch
@curryable
def reduce(self:L, f, initial=None):
    "Wrapper for `functools.reduce`"
    return reduce(f, self) if initial is None else reduce(f, self, initial)

# %% ../nbs/02_foundation.ipynb #3db27a6a
@patch
@curryable
def starreduce(self:L, f, initial=None):
    "Like `reduce`, but unpacks elements as args to `f`"
    _f = lambda acc, x: f(acc, *x)
    return reduce(_f, self) if initial is None else reduce(_f, self, initial)

# %% ../nbs/02_foundation.ipynb #046f01fb
@patch
@curryable
def rstarreduce(self:L, f, initial=None):
    "Like `starreduce`, but reverse the order of unpacked args"
    _f = lambda acc, x: f(acc, *x[::-1])
    return reduce(_f, self) if initial is None else reduce(_f, self, initial)

# %% ../nbs/02_foundation.ipynb #84b3413b
@patch
def sum(self:L):
    "Sum of the items"
    return self.reduce(operator.add, 0)

# %% ../nbs/02_foundation.ipynb #2a96a0fd
@patch
def product(self:L):
    "Product of the items"
    return self.reduce(operator.mul, 1)

# %% ../nbs/02_foundation.ipynb #f82d9975
@patch
def map_first(self:L, f=noop, g=noop, *args, **kwargs):
    "First element of `map_filter`"
    return first(self.map(f, *args, **kwargs), g)

# %% ../nbs/02_foundation.ipynb #1366f0fa
@patch
def setattrs(self:L, attr, val):
    "Call `setattr` on all items"
    [setattr(o,attr,val) for o in self]

# %% ../nbs/02_foundation.ipynb #ab6bb396
@patch
def flatmap(self:L, f, **kwargs):
    "Apply f to each element and flatten the results into a single L."
    return L(flatmap(f, self, **kwargs))

# %% ../nbs/02_foundation.ipynb #9c927236
@patch
def cycle(self:L):
    "Same as `itertools.cycle`"
    return cycle(self)

# %% ../nbs/02_foundation.ipynb #79e84e6d
@patch
@curryable
def takewhile(self:L, f):
    "Same as `itertools.takewhile`"
    return self._new(itertools.takewhile(f, self))

# %% ../nbs/02_foundation.ipynb #c9f6e354
@patch
@curryable
def dropwhile(self:L, f):
    "Same as `itertools.dropwhile`"
    return self._new(itertools.dropwhile(f, self))

# %% ../nbs/02_foundation.ipynb #45f04c5c
@patch
@curryable
def startakewhile(self:L, f):
    "Like `takewhile`, but unpacks elements as args to `f`"
    return self._new(itertools.takewhile(lambda x: f(*x), self))

# %% ../nbs/02_foundation.ipynb #f64504c1
@patch
@curryable
def rstartakewhile(self:L, f):
    "Like `startakewhile`, but reverse the order of args"
    return self._new(itertools.takewhile(lambda x: f(*x[::-1]), self))

# %% ../nbs/02_foundation.ipynb #6544466c
@patch
@curryable
def stardropwhile(self:L, f):
    "Like `dropwhile`, but unpacks elements as args to `f`"
    return self._new(itertools.dropwhile(lambda x: f(*x), self))

# %% ../nbs/02_foundation.ipynb #91f75f7d
@patch
@curryable
def rstardropwhile(self:L, f):
    "Like `stardropwhile`, but reverse the order of args"
    return self._new(itertools.dropwhile(lambda x: f(*x[::-1]), self))

# %% ../nbs/02_foundation.ipynb #53cf2acc
@patch
@curryable
def accumulate(self:L, f=operator.add, initial=None):
    "Same as `itertools.accumulate`"
    return self._new(itertools.accumulate(self, f, initial=initial))

# %% ../nbs/02_foundation.ipynb #c55df328
@patch
def pairwise(self:L):
    "Same as `itertools.pairwise`"
    return self._new(itertools.pairwise(self))

# %% ../nbs/02_foundation.ipynb #1f2a370c
def _batched(iterable, n):
    "Batch data into tuples of length n. The last batch may be shorter."
    if n < 1: raise ValueError('n must be at least one')
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch

# %% ../nbs/02_foundation.ipynb #41aefe58
try: from itertools import batched
except ImportError: batched = _batched

# %% ../nbs/02_foundation.ipynb #d9f7ffb9
@patch
def batched(self:L, n):
    "Same as `itertools.batched` (but also works on older Python versions"
    return self._new(batched(self, n))

# %% ../nbs/02_foundation.ipynb #9e996aec
@patch
def compress(self:L, selectors):
    "Same as `itertools.compress`"
    return self._new(itertools.compress(self, selectors))

# %% ../nbs/02_foundation.ipynb #051afac0
@patch
def permutations(self:L, r=None):
    "Same as `itertools.permutations`"
    return self._new(itertools.permutations(self, r))

# %% ../nbs/02_foundation.ipynb #63066166
@patch
def combinations(self:L, r):
    "Same as `itertools.combinations`"
    return self._new(itertools.combinations(self, r))

# %% ../nbs/02_foundation.ipynb #f29ba907
@patch
@curryable
def partition(self:L, f=noop, **kwargs):
    "Split into two `L`s based on predicate `f`: (true_items, false_items)"
    a,b = [],[]
    for o in self: (a if f(o, **kwargs) else b).append(o)
    return self._new(a),self._new(b)

# %% ../nbs/02_foundation.ipynb #2dd5af2f
@patch
@curryable
def starpartition(self:L, f, **kwargs):
    "Like `partition`, but unpacks elements as args to `f`"
    a,b = [],[]
    for o in self: (a if f(*o, **kwargs) else b).append(o)
    return self._new(a),self._new(b)

# %% ../nbs/02_foundation.ipynb #63c70884
@patch
@curryable
def rstarpartition(self:L, f, **kwargs):
    "Like `starpartition`, but reverse the order of args"
    a,b = [],[]
    for o in self: (a if f(*o[::-1], **kwargs) else b).append(o)
    return self._new(a),self._new(b)

# %% ../nbs/02_foundation.ipynb #3c269425
@patch
def flatten(self:L):
    "Recursively flatten nested iterables (except strings)"
    def _flatten(o):
        for item in o:
            if isinstance(item, (str,bytes)) or not hasattr(item,'__iter__'): yield item
            else: yield from _flatten(item)
    return self._new(_flatten(self))

# %% ../nbs/02_foundation.ipynb #8e2bafde
def save_config_file(file, d, **kwargs):
    "Write settings dict to a new config file, or overwrite the existing one."
    config = ConfigParser(**kwargs)
    config['DEFAULT'] = d
    config.write(open(file, 'w'))

# %% ../nbs/02_foundation.ipynb #08333fdc
def read_config_file(file, **kwargs):
    config = ConfigParser(**kwargs)
    config.read(file, encoding='utf8')
    return config['DEFAULT']

# %% ../nbs/02_foundation.ipynb #95834369
def find_file_parents(fname, frompath=None):
    "Search `cfg_path` and its parents to find `cfg_name`"
    p = Path(frompath or Path.cwd()).expanduser().absolute()
    return first(o for o in [p, *p.parents] if (o/fname).exists())

# %% ../nbs/02_foundation.ipynb #e06640e8
class Config:
    "Reading and writing `ConfigParser` ini files"
    def __init__(self, cfg_path, cfg_name, create=None, save=True, extra_files=None, types=None, **cfg_kwargs):
        self.types = types or {}
        cfg_path = Path(cfg_path).expanduser().absolute()
        self.config_path,self.config_file = cfg_path,cfg_path/cfg_name
        self._cfg = ConfigParser(**cfg_kwargs)
        self.d = self._cfg['DEFAULT']
        found = [Path(o) for o in self._cfg.read(L(extra_files)+[self.config_file], encoding='utf8')]
        if self.config_file not in found and create is not None:
            self._cfg.read_dict({'DEFAULT':create})
            if save:
                cfg_path.mkdir(exist_ok=True, parents=True)
                save_config_file(self.config_file, create)

    def __repr__(self): return repr(dict(self._cfg.items('DEFAULT', raw=True)))
    def __setitem__(self,k,v): self.d[k] = str(v)
    def __contains__(self,k):  return k in self.d
    def save(self):            save_config_file(self.config_file,self.d)
    def __getattr__(self,k):   return stop(AttributeError(k)) if k=='d' or k not in self.d else self.get(k)
    def __getitem__(self,k):   return stop(IndexError(k)) if k not in self.d else self.get(k)

    def get(self,k,default=None):
        v = self.d.get(k, default)
        if v is None: return None
        typ = self.types.get(k, None)
        if typ==bool: return str2bool(v)
        if not typ: return str(v)
        if typ==Path: return self.config_path/v
        return typ(v)

    def path(self,k,default=None):
        v = self.get(k, default)
        return v if v is None else self.config_path/v

    @classmethod
    def find(cls, cfg_name, cfg_path=None, **kwargs):
        "Search `cfg_path` and its parents to find `cfg_name`"
        p = find_file_parents(cfg_name, cfg_path)
        return cls(p, cfg_name, **kwargs) if p else None



"""Helpful tools for running cli commands and reading, modifying, and creating files in python. This is used primarily for AI's in tool loops for automating tasks involving the filesystem."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/12_tools.ipynb.

# %% auto #0
__all__ = ['valid_paths', 'explain_exc', 'ensure', 'valid_path', 'run_cmd', 'rg', 'sed', 'view', 'create', 'insert',
           'str_replace', 'strs_replace', 'replace_lines', 'move_lines', 'get_callable']

# %% ../nbs/12_tools.ipynb #578246d2
from .xdg import *
from .imports import *
from .xtras import truncstr
from shlex import split
from subprocess import run, DEVNULL

# %% ../nbs/12_tools.ipynb #7603dedb
def explain_exc(task=''):
    """Convert an current exception to  an LLM friendly error message."""
    try: raise sys.exc_info()[1]
    except (AssertionError, ZeroDivisionError, ValueError, FileNotFoundError) as e:
        return f"Error: {e}"
    except Exception as e: return f"Error {task}: {repr(e)}"


# %% ../nbs/12_tools.ipynb #b8823d07
def ensure(b: bool, msg:str=""):
    "Works like assert b, msg but raise ValueError and is not disabled when run with python -O"
    if not b: raise ValueError(msg)


# %% ../nbs/12_tools.ipynb #6afaf015
def _load_valid_paths():
    cfg = xdg_config_home() / 'fc_tools_paths'
    base = ['.', '/tmp']
    if not cfg.exists(): return base
    return base + cfg.read_text().split()

valid_paths = _load_valid_paths()

# %% ../nbs/12_tools.ipynb #e4288129
def valid_path(path:str, must_exist:bool=True, chk_perms:bool=True) -> Path:
    'Return expanded/resolved Path, raising FileNotFoundError if must_exist and missing'
    p = Path(path).expanduser().resolve()
    vpaths = [Path(vp).expanduser().resolve() for vp in valid_paths]
    if chk_perms and not any(p == vp or vp in p.parents for vp in vpaths): raise PermissionError(f'Path not in valid_paths: {p}')
    if must_exist and not p.exists(): raise FileNotFoundError(f'File not found: {p}')
    return p

# %% ../nbs/12_tools.ipynb #1182336d
def run_cmd(
    cmd:str, # The command name to run
    argstr:str='', # All args to the command, will be split with shlex
    disallow_re:str=None, # optional regex which, if matched on argstr, will disallow the command
    allow_re:str=None # optional regex which, if not matched on argstr, will disallow the command
):
    "Run `cmd` passing split `argstr`, optionally checking for allowed argstr"
    try:
        ensure(not (disallow_re and re.search(disallow_re, argstr)), 'args disallowed')
        ensure(not (allow_re    and re.search(   allow_re, argstr)), 'args not allowed')
        argstr = argstr.replace('\\', '\\\\')
        args = [str(Path(a).expanduser()) if a.startswith('~') else a for a in split(argstr, posix=True)]
        outp = run([cmd] + args, text=True, stdin=DEVNULL, capture_output=True)
    except: return explain_exc(f'running cmd')
    res = outp.stdout
    if res and outp.stderr: res += '\n'
    return res + outp.stderr

# %% ../nbs/12_tools.ipynb #eb253a39
@llmtool
def rg(
    argstr:str, # All args to the command, will be split with shlex
    disallow_re:str=None, # optional regex which, if matched on argstr, will disallow the command
    allow_re:str=None # optional regex which, if not matched on argstr, will disallow the command
):
    "Run the `rg` command with the args in `argstr`"
    return run_cmd('rg', '-n '+argstr, disallow_re=disallow_re, allow_re=allow_re)

# %% ../nbs/12_tools.ipynb #09de7b32
@llmtool
def sed(
    argstr:str, # All args to the command, will be split with shlex
    disallow_re:str=None, # optional regex which, if matched on argstr, will disallow the command
    allow_re:str=None # optional regex which, if not matched on argstr, will disallow the command
):
    "Run the `sed` command with the args in `argstr` (e.g for reading a section of a file)"
    return run_cmd('sed', argstr, allow_re=allow_re, disallow_re=disallow_re)

# %% ../nbs/12_tools.ipynb #864b213a
def _fmt_path(f, p, skip_folders=()):
    'Format path with emoji for dirs/symlinks or size for files'
    parts = f.relative_to(p).parts
    if any(part.startswith('.') for part in parts): return None
    if any(part in skip_folders for part in parts): return None
    if f.is_symlink(): return f'{f} 🔗'
    if f.is_dir(): return f'{f} 📁'
    return f'{f} ({f.stat().st_size/1024:.1f}k)'

# %% ../nbs/12_tools.ipynb #5dd1aaf3
@llmtool
def view(
    path:str, # Path to directory or file to view
    view_range:tuple[int,int]=None, # Optional 1-indexed (start, end) line range for files, end=-1 for EOF. Do NOT use unless it's known that the file is too big to keep in context—simply view the WHOLE file when possible
    nums:bool=False, # Whether to show line numbers
    skip_folders:tuple[str,...]=('_proc','__pycache__') # Folder names to skip when listing directories
):
    'View directory or file contents with optional line range and numbers'
    try:
        p = valid_path(path, chk_perms=False)
        header = None
        if p.is_dir():
            lines = [s for f in p.glob('**/*') if (s := _fmt_path(f, p, skip_folders))]
            header = f'Directory contents of {p}:'
        else: lines = p.read_text().splitlines()
        s, e = 1, len(lines)
        if view_range:
            s,e = view_range
            ensure(1<=s<=len(lines), f'Invalid start line {s}')
            ensure(e==-1 or s<=e<=len(lines), f'Invalid end line {e}')
            lines = lines[s-1:None if e==-1 else e]
        if nums: lines = [f'{i+s:6d} │ {l}' for i, l in enumerate(lines)]
        content = '\n'.join(lines)
        return f'{header}\n{content}' if header else content
    except: return explain_exc('viewing')

# %% ../nbs/12_tools.ipynb #36f58e38
@llmtool
def create(
    path: str, # Path where the new file should be created
    file_text: str, # Content to write to the file
    overwrite:bool=False # Whether to overwrite existing files
) -> str:
    'Creates a new file with the given content at the specified path'
    try:
        p = valid_path(path, must_exist=False)
        if p.exists():
            if not overwrite: return f'Error: File already exists: {p}'
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_text(file_text)
        return f'Created file {p}.'
    except: return explain_exc('creating file')

# %% ../nbs/12_tools.ipynb #434147ef
@llmtool
def insert(
    path: str, # Path to the file to modify
    insert_line: int, # Line number where to insert (0-based indexing)
    new_str: str # Text to insert at the specified line
) -> str:
    'Insert new_str at specified line number'
    try:
        p = valid_path(path)
        content = p.read_text().splitlines()
        ensure(0<=insert_line<=len(content), f'Invalid line number {insert_line}')
        content.insert(insert_line, new_str)
        new_content = '\n'.join(content)
        p.write_text(new_content)
        return f'Inserted text at line {insert_line} in {p}'
    except: return explain_exc('inserting text')

# %% ../nbs/12_tools.ipynb #9272ff7d
@llmtool
def str_replace(
    path: str, # Path to the file to modify
    old_str: str, # Text to find and replace
    new_str: str # Text to replace with
) -> str:
    'Replace first occurrence of old_str with new_str in file'
    try:
        p = valid_path(path)
        content = p.read_text()
        count = content.count(old_str)
        if count == 0: return f'Error: Text "{truncstr(old_str, 10)}" not found in file'
        if count > 1: return f'Error: Multiple matches found ({count}) of "{truncstr(old_str, 10)}"'
        new_content = content.replace(old_str, new_str, 1)
        p.write_text(new_content)
        return f'Replaced text in {p}'
    except: return explain_exc('replacing text')

# %% ../nbs/12_tools.ipynb #eb907119
@llmtool
def strs_replace(
    path:str, # Path to the file to modify
    old_strs:list[str], # List of strings to find and replace
    new_strs:list[str], # List of replacement strings (must match length of old_strs)
):
    "Replace for each str pair in old_strs,new_strs"
    res = [str_replace(path, old, new) for (old,new) in zip(old_strs,new_strs)]
    return 'Results for each replacement:\n' + '; '.join(res)

# %% ../nbs/12_tools.ipynb #94dd09ed
@llmtool
def replace_lines(
    path:str, # Path to the file to modify
    start_line:int, # Starting line number to replace (1-based indexing)
    end_line:int, # Ending line number to replace (1-based indexing, inclusive)
    new_content:str, # New content to replace the specified lines
):
    "Replace lines in file using start and end line-numbers (index starting at 1)"
    try:
        p = valid_path(path)
        content = p.readlines()
        if not new_content.endswith('\n'): new_content+='\n'
        content[start_line-1:end_line] = [new_content]
        p.write_text(''.join(content))
        return f"Replaced lines {start_line} to {end_line}."
    except: return explain_exc('replacing lines')

# %% ../nbs/12_tools.ipynb #b136abf8
@llmtool
def move_lines(
    path: str,  # Path to the file to modify
    start_line: int,  # Starting line number to move (1-based)
    end_line: int,  # Ending line number to move (1-based, inclusive)
    dest_line: int,  # Destination line number (1-based, where lines will be inserted before)
) -> str:
    "Move lines from start_line:end_line to before dest_line"
    try:
        p = valid_path(path)
        lines = p.read_text().splitlines()
        ensure(1 <= start_line <= end_line <= len(lines), f"Invalid range {start_line}-{end_line}")
        ensure(1 <= dest_line <= len(lines) + 1, f"Invalid destination {dest_line}")
        ensure(not(start_line <= dest_line <= end_line + 1), "Destination within source range")
        
        chunk = lines[start_line-1:end_line]
        del lines[start_line-1:end_line]
        # Adjust dest if it was after the removed chunk
        if dest_line > end_line: dest_line -= len(chunk)
        lines[dest_line-1:dest_line-1] = chunk
        p.write_text('\n'.join(lines) + '\n')
        return f"Moved lines {start_line}-{end_line} to line {dest_line}"
    except: return explain_exc()

# %% ../nbs/12_tools.ipynb #fc53b116
def get_callable():
    "Return callable objects defined in caller's module"
    import inspect
    g = inspect.currentframe().f_back.f_globals
    return {
        f:o for f,o in g.items()
        if callable(o) and hasattr(o, '__module__') and o.__module__ == '__main__' and not f.startswith('_')
    }



"""A fast way to turn your python function into a script."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_script.ipynb.

# %% auto #0
__all__ = ['SCRIPT_INFO', 'store_true', 'store_false', 'bool_arg', 'clean_type_str', 'Param', 'anno_parser', 'args_from_prog',
           'set_ctx', 'call_parse']

# %% ../nbs/06_script.ipynb #8a36db98
import inspect,argparse,shutil,types

from functools import wraps,partial
from .imports import *
from .utils import *
from .docments import docments
from typing import get_origin, get_args, Union

# %% ../nbs/06_script.ipynb #5bf7ac6c
def store_true():
    "Placeholder to pass to `Param` for `store_true` action"
    pass

# %% ../nbs/06_script.ipynb #e2798370
def store_false():
    "Placeholder to pass to `Param` for `store_false` action"
    pass

# %% ../nbs/06_script.ipynb #69d525bc
def bool_arg(v):
    "Use as `type` for `Param` to get `bool` behavior"
    return str2bool(v)

# %% ../nbs/06_script.ipynb #f1d4a6a3
def clean_type_str(x:str):
    x = str(x)
    x = re.sub(r"(enum |class|function|__main__\.|\ at.*)", '', x)
    x = re.sub(r"(<|>|'|\ )", '', x) # spl characters
    return x

# %% ../nbs/06_script.ipynb #6785a783
class Param:
    "A parameter in a function used in `anno_parser` or `call_parse`"
    def __init__(self, help="", type=None, opt=True, action=None, nargs=None, const=None,
                 choices=None, required=None, default=None, version=None):
        if type==store_true:  type,action,default=None,'store_true',False
        if type==store_false: type,action,default=None,'store_false',True
        if type and isinstance(type,typing.Type) and issubclass(type,enum.Enum) and not choices: choices=list(type)
        help = help or ""
        self.negated = False
        store_attr()

    def set_default(self, d):
        if self.action == "version":
            if d != inspect.Parameter.empty: self.version = d
            self.opt = True
            return
        if self.type==bool and self.action is None:
            self.type = None
            if d is True: self.action,self.default,self.negated = 'store_false',True,True
            else:         self.action,self.default = 'store_true',False
        elif self.default is None:
            if d == inspect.Parameter.empty: self.opt = False
            else: self.default = d
        if self.default is not None: self.help += f" (default: {self.default})"

    @property
    def pre(self): return '--' if self.opt else ''
    @property
    def kwargs(self): return {k:v for k,v in self.__dict__.items() if v is not None and k not in ('opt','negated') and k[0]!='_'}
    def __repr__(self):
        if not self.help and self.type is None: return ""
        if not self.help and self.type is not None: return f"{clean_type_str(self.type)}"
        if self.help and self.type is None: return f"<{self.help}>"
        if self.help and self.type is not None: return f"{clean_type_str(self.type)} <{self.help}>"

# %% ../nbs/06_script.ipynb #cceb4486
class _HelpFormatter(argparse.HelpFormatter):
    def __init__(self, prog, indent_increment=2):
        cols = shutil.get_terminal_size((120,30))[0]
        super().__init__(prog, max_help_position=cols//2, width=cols, indent_increment=indent_increment)
    def _expand_help(self, action): return self._get_help_string(action)

# %% ../nbs/06_script.ipynb #2259c093
def _is_union(t): return get_origin(t) in (Union, types.UnionType) if hasattr(types, 'UnionType') else get_origin(t) is Union

def _union_parser(types):
    "Return a parser that tries each type in sequence"
    def _parse(v):
        for t in types:
            if t is type(None): continue
            try: return t(v)
            except: pass
        raise ValueError(f"Could not parse {v!r} as any of {types}")
    return _parse

def _union_type(t):
    "Get parser for Union types, or None if not a Union"
    if not _is_union(t): return None
    return _union_parser(get_args(t))

# %% ../nbs/06_script.ipynb #5e5bea67
def anno_parser(func, prog:str=None):
    "Look at params (annotated with `Param`) in func and return an `ArgumentParser`"
    p = argparse.ArgumentParser(description=func.__doc__, prog=prog, formatter_class=_HelpFormatter)
    for k,v in docments(func, full=True, returns=False, eval_str=True).items():
        param = v.anno
        if not isinstance(param, Param):
            anno = _union_type(v.anno) or v.anno
            param = Param(v.docment, anno)
        param.set_default(v.default)
        name = f"no-{k}" if param.negated else k
        kw = param.kwargs
        if param.negated: kw['dest'] = k
        p.add_argument(f"{param.pre}{name}", **kw)
    p.add_argument(f"--pdb", help=argparse.SUPPRESS, action='store_true')
    p.add_argument(f"--xtra", help=argparse.SUPPRESS, type=str)
    return p

# %% ../nbs/06_script.ipynb #75e8a419
def args_from_prog(func, prog):
    "Extract args from `prog`"
    if prog is None or '#' not in prog: return {}
    if '##' in prog: _,prog = prog.split('##', 1)
    progsp = prog.split("#")
    args = {progsp[i]:progsp[i+1] for i in range(0, len(progsp), 2)}
    annos = type_hints(func)
    for k,v in args.items():
        anno = annos.get(k)
        t = getattr(anno, 'type', anno)
        if t in (bool, bool_arg): t = str2bool
        elif isinstance(anno, Param) and anno.action in ('store_true', 'store_false'): t = str2bool
        if t: args[k] = t(v)
    return args

# %% ../nbs/06_script.ipynb #f76b07f6
SCRIPT_INFO = SimpleNamespace(func=None)

# %% ../nbs/06_script.ipynb #42c8e85f
from contextvars import ContextVar
from contextlib import contextmanager

# %% ../nbs/06_script.ipynb #e4537112
@contextmanager
def set_ctx(cv, val=True):
    token = cv.set(val)
    try: yield
    finally: cv.reset(token)

# %% ../nbs/06_script.ipynb #fc816498
_in_call_parse = ContextVar('_in_call_parse', default=False)

def call_parse(func=None, nested=False):
    "Decorator to create a simple CLI from `func` using `anno_parser`"
    if func is None: return partial(call_parse, nested=nested)

    @wraps(func)
    def _f(*args, **kwargs):
        if args or kwargs or _in_call_parse.get(): return func(*args, **kwargs)
        with set_ctx(_in_call_parse):
            mod = inspect.getmodule(inspect.currentframe().f_back)
            if not mod: return func(*args, **kwargs)
            if not SCRIPT_INFO.func and mod.__name__=="__main__": SCRIPT_INFO.func = func.__name__
            if len(sys.argv)>1 and sys.argv[1]=='': sys.argv.pop(1)
            p = anno_parser(func)
            if nested: args, sys.argv[1:] = p.parse_known_args()
            else: args = p.parse_args()
            args = args.__dict__
            xtra = otherwise(args.pop('xtra', ''), eq(1), p.prog)
            tfunc = trace(func) if args.pop('pdb', False) else func
            return tfunc(**merge(args, args_from_prog(func, xtra)))

    mod = inspect.getmodule(inspect.currentframe().f_back)
    if getattr(mod, '__name__', '') =="__main__":
        setattr(mod, func.__name__, _f)
        SCRIPT_INFO.func = func.__name__
        return _f()
    else: return _f



"""Document parameters using comments."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_docments.ipynb.

# %% auto #0
__all__ = ['empty', 'docstring', 'parse_docstring', 'isdataclass', 'get_dataclass_source', 'get_source', 'get_name', 'qual_name',
           'docments', 'sig_source', 'extract_docstrings', 'DocmentTbl', 'DocmentList', 'DocmentText', 'sig2str',
           'ShowDocRenderer', 'MarkdownRenderer']

# %% ../nbs/04_docments.ipynb #4c662acc
import re,ast,inspect
from tokenize import tokenize,COMMENT
from ast import parse,FunctionDef,AsyncFunctionDef,AnnAssign
from io import BytesIO
from textwrap import dedent
from types import SimpleNamespace
from inspect import getsource,isfunction,ismethod,isclass,signature,Parameter
from dataclasses import dataclass, is_dataclass
from .utils import *
from .meta import delegates
from . import docscrape
from textwrap import fill
from inspect import isclass,getdoc,signature

# %% ../nbs/04_docments.ipynb #53e21187
def docstring(sym):
    "Get docstring for `sym` for functions ad classes"
    if isinstance(sym, str): return sym
    res = getdoc(sym)
    if not res and isclass(sym): res = getdoc(sym.__init__)
    return res or ""

# %% ../nbs/04_docments.ipynb #1e0cf854
def parse_docstring(sym):
    "Parse a numpy-style docstring in `sym`"
    return AttrDict(**docscrape.NumpyDocString(docstring(sym)))

# %% ../nbs/04_docments.ipynb #b1a586d1
def isdataclass(s):
    "Check if `s` is a dataclass but not a dataclass' instance"
    return is_dataclass(s) and isclass(s)

# %% ../nbs/04_docments.ipynb #6549c4b2
def get_dataclass_source(s):
    "Get source code for dataclass `s`"
    return getsource(s) if not getattr(s, "__module__") == '__main__' else ""

# %% ../nbs/04_docments.ipynb #cac59989
def get_source(s):
    "Get source code for string, function object or dataclass `s`"
    if isinstance(s,str): return s
    return getsource(s) if isfunction(s) or ismethod(s) else get_dataclass_source(s) if isdataclass(s) else None

# %% ../nbs/04_docments.ipynb #91c0d15f
def _parses(s):
    "Parse Python code in string, function object or dataclass `s`"
    return parse(dedent(get_source(s) or ''))

def _tokens(s):
    "Tokenize Python code in string or function object `s`"
    s = get_source(s)
    if not s: return []
    return tokenize(BytesIO(s.encode('utf-8')).readline)

_clean_re = re.compile(r'^\s*#(.*)\s*$')
def _clean_comment(s):
    res = _clean_re.findall(s)
    return res[0] if res else None

def _param_locs(s, returns=True, args_kwargs=False):
    "`dict` of parameter line numbers to names"
    body = _parses(s).body
    if len(body)==1:
        defn = body[0]
        if isinstance(defn, (FunctionDef, AsyncFunctionDef)):
            res = {arg.lineno:arg.arg for arg in defn.args.args}
            # Add *args if present
            if defn.args.vararg: res[defn.args.vararg.lineno] = defn.args.vararg.arg
            # Add keyword-only args
            res.update({arg.lineno:arg.arg for arg in defn.args.kwonlyargs})
            # Add **kwargs if present
            if defn.args.kwarg and args_kwargs: res[defn.args.kwarg.lineno] = defn.args.kwarg.arg
            if returns and defn.returns: res[defn.returns.lineno] = 'return'
            return res
        elif isdataclass(s):
            res = {arg.lineno:arg.target.id for arg in defn.body if isinstance(arg, AnnAssign)}
            return res
    return None

# %% ../nbs/04_docments.ipynb #93930c1f
empty = Parameter.empty

# %% ../nbs/04_docments.ipynb #6434e80e
def _get_comment(line, arg, comments, parms):
    if line in comments: return comments[line].strip()
    line -= 1
    res = []
    while line and line in comments and line not in parms:
        res.append(comments[line])
        line -= 1
    return dedent('\n'.join(reversed(res))) if res else None

def _get_full(p, docs, eval_str=False):
    anno = p.annotation
    if anno==empty:
        if p.default!=empty: anno = type(p.default)
        elif p.kind in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD): anno = p.kind
        elif eval_str: anno = None
    return AttrDict(docment=docs.get(p.name), anno=anno, default=p.default)

# %% ../nbs/04_docments.ipynb #1b4d817c
def _merge_doc(dm, npdoc):
    if not npdoc: return dm
    if not isinstance(dm, dict): return dm or '\n'.join(npdoc.desc)
    # if not dm.anno or dm.anno==empty: dm.anno = npdoc.type
    if not dm.docment: dm.docment = '\n'.join(npdoc.desc)
    return dm

def _merge_docs(dms, npdocs):
    npparams = npdocs['Parameters']
    params = {nm:_merge_doc(dm,npparams.get(nm,None)) for nm,dm in dms.items()}
    if 'return' in dms: params['return'] = _merge_doc(dms['return'], npdocs['Returns'])
    return params

# %% ../nbs/04_docments.ipynb #40e3f274
def _get_property_name(p):
    "Get the name of property `p`"
    if hasattr(p, 'fget'):
        return p.fget.func.__qualname__ if hasattr(p.fget, 'func') else p.fget.__qualname__
    else: return next(iter(re.findall(r'\'(.*)\'', str(p)))).split('.')[-1]

# %% ../nbs/04_docments.ipynb #da0465e3
def get_name(obj):
    "Get the name of `obj`"
    if isinstance(obj, partial):
        nm = get_name(obj.func)
        args = [repr(a) for a in obj.args] + [f'{k}={repr(v)}' for k,v in obj.keywords.items()]
        return f"{nm}[partial: {', '.join(args)}]"
    if hasattr(obj, '__name__'):       return obj.__name__
    elif getattr(obj, '_name', False): return obj._name
    elif hasattr(obj,'__origin__'):    return str(obj.__origin__).split('.')[-1]
    elif type(obj)==property:          return _get_property_name(obj)
    else:                              return str(obj).split('.')[-1]

# %% ../nbs/04_docments.ipynb #5e273f22
def qual_name(obj):
    "Get the qualified name of `obj`"
    if hasattr(obj,'__qualname__'): return obj.__qualname__
    if ismethod(obj):       return f"{get_name(obj.__self__)}.{get_name(fn)}"
    return get_name(obj)

# %% ../nbs/04_docments.ipynb #9b62ab20
def docments(s, full=False, eval_str=False, returns=True, args_kwargs=False):
    "Get docments for `s`"
    if isclass(s) and not is_dataclass(s): s = s.__init__
    try: sig = signature_ex(s, eval_str=eval_str)
    except ValueError: return AttrDict()
    nps = parse_docstring(s)
    docs = {}
    while s:
        p = _param_locs(s, returns=returns, args_kwargs=args_kwargs) or {}
        c = {o.start[0]:_clean_comment(o.string) for o in _tokens(s) if o.type==COMMENT}
        for k,v in p.items():
            if v not in docs: docs[v] = _get_comment(k, v, c, p)
        s = getattr(s, '__delwrap__', None)
    
    res = {k:_get_full(v, docs, eval_str=eval_str) if full else docs.get(k) for k,v in sig.parameters.items()}
    if returns:
        if full: res['return'] = AttrDict(docment=docs.get('return'), anno=sig.return_annotation, default=empty)
        else: res['return'] = docs.get('return')
    return AttrDict(_merge_docs(res, nps))

# %% ../nbs/04_docments.ipynb #40cdbeb2
def sig_source(obj):
    "Full source of signature line(s) for a function or class."
    src = inspect.getsource(obj)
    tree = ast.parse(src)
    body_start = tree.body[0].body[0].lineno
    if body_start == 1: return src.splitlines()[0]
    return '\n'.join(src.splitlines()[:body_start-1])

# %% ../nbs/04_docments.ipynb #6f0d6ea4
def _get_params(node):
    params = [a.arg for a in node.args.args]
    if node.args.vararg: params.append(f"*{node.args.vararg.arg}")
    if node.args.kwarg: params.append(f"**{node.args.kwarg.arg}")
    return ", ".join(params)

# %% ../nbs/04_docments.ipynb #180b4c2d
class _DocstringExtractor(ast.NodeVisitor):
    def __init__(self): self.docstrings,self.cls,self.cls_init = {},None,None

    def visit_FunctionDef(self, node):
        name = node.name
        if name == '__init__':
            self.cls_init = node
            return
        elif name.startswith('_'): return
        elif self.cls: name = f"{self.cls}.{node.name}"
        docs = ast.get_docstring(node)
        params = _get_params(node)
        if docs: self.docstrings[name] = (docs, params)
        self.generic_visit(node)

    def visit_ClassDef(self, node):
        self.cls,self.cls_init = node.name,None
        docs = ast.get_docstring(node)
        if docs: self.docstrings[node.name] = ()
        self.generic_visit(node)
        if not docs and self.cls_init: docs = ast.get_docstring(self.cls_init)
        params = _get_params(self.cls_init) if self.cls_init else ""
        if docs: self.docstrings[node.name] = (docs, params)
        self.cls,self.cls_init = None,None

    def visit_Module(self, node):
        module_doc = ast.get_docstring(node)
        if module_doc: self.docstrings['_module'] = (module_doc, "")
        self.generic_visit(node)

# %% ../nbs/04_docments.ipynb #b1d612e9
def extract_docstrings(code):
    "Create a dict from function/class/method names to tuples of docstrings and param lists"
    extractor = _DocstringExtractor()
    extractor.visit(ast.parse(code))
    return extractor.docstrings

# %% ../nbs/04_docments.ipynb #e4e1b815
def _non_empty_keys(d:dict): return L([k for k,v in d.items() if v != inspect._empty])
def _bold(s): return f'**{s}**' if s.strip() else s

# %% ../nbs/04_docments.ipynb #ac070254
def _escape_markdown(s):
    for c in '|^': s = re.sub(rf'\\?\{c}', rf'\{c}', s)
    return s.replace('\n', '<br>')

# %% ../nbs/04_docments.ipynb #ced78f56
def _maybe_nm(o):
    if (o == inspect._empty): return ''
    else: return o.__name__ if hasattr(o, '__name__') else str(o)

# %% ../nbs/04_docments.ipynb #fe6d83f1
def _list2row(l:list): return '| '+' | '.join([_maybe_nm(o) for o in l]) + ' |'

# %% ../nbs/04_docments.ipynb #44ac2e4f
class _DocmentBase:
    def __init__(self, obj):
        self.obj,self.dm = obj, docments(obj, full=True)
        if 'self' in self.dm: del self.dm['self']
    
    @property
    def has_docment(self): return any(v.get('docment') for v in self.dm.values())

# %% ../nbs/04_docments.ipynb #b45f731d
class DocmentTbl(_DocmentBase):
    _map = {'anno':'Type', 'default':'Default', 'docment':'Details'}

    def __init__(self, obj, verbose=True, returns=True):
        "Compute the docment table string"
        super().__init__(obj)
        self.verbose = verbose
        self.returns = False if isdataclass(obj) else returns
        try: self.params = L(signature(obj, eval_str=True).parameters.keys())
        except (ValueError,TypeError): self.params=[]
        for d in self.dm.values(): d['docment'] = ifnone(d['docment'], inspect._empty)

    @property
    def _columns(self):
        "Compute the set of fields that have at least one non-empty value so we don't show tables empty columns"
        cols = set(flatten(L(self.dm.values()).filter().map(_non_empty_keys)))
        candidates = self._map if self.verbose else {'docment': 'Details'}
        return {k:v for k,v in candidates.items() if k in cols}

    @property
    def has_docment(self): return 'docment' in self._columns and self._row_list

    @property
    def has_return(self): return self.returns and bool(_non_empty_keys(self.dm.get('return', {})))

    def _row(self, nm, props): return [nm] + [props[c] for c in self._columns]

    @property
    def _row_list(self):
        ordered_params = [(p, self.dm[p]) for p in self.params if p != 'self' and p in self.dm]
        return L([self._row(nm, props) for nm,props in ordered_params])

    @property
    def _hdr_list(self): return ['  '] + [_bold(l) for l in L(self._columns.values())]

    @property
    def hdr_str(self):
        md = _list2row(self._hdr_list)
        return md + '\n' + _list2row(['-' * len(l) for l in self._hdr_list])

    @property
    def params_str(self): return '\n'.join(self._row_list.map(_list2row))

    @property
    def return_str(self): return _list2row(['**Returns**']+[_bold(_maybe_nm(self.dm['return'][c])) for c in self._columns])

    def _repr_markdown_(self):
        if not self.has_docment: return ''
        _tbl = [self.hdr_str, self.params_str]
        if self.has_return: _tbl.append(self.return_str)
        return '\n'.join(_tbl)

    def __eq__(self,other): return self.__str__() == str(other).strip()
    __str__ = _repr_markdown_
    __repr__ = basic_repr()

# %% ../nbs/04_docments.ipynb #5af61ab1
class DocmentList(_DocmentBase):
    def _fmt(self, nm, p):
        anno,default,doc = _maybe_nm(p.get('anno','')), p.get('default',empty), p.get('docment','')
        s = f'`{nm}{":" + anno if anno else ""}{"=" + _maybe_nm(default) if default != empty else ""}`'
        br = '\xa0'*3
        return f'- {s}' + (f'{br}*{doc}*' if doc else '')

    def _repr_markdown_(self): return '\n'.join(self._fmt(k,v) for k,v in self.dm.items())
    __repr__=__str__=_repr_markdown_

# %% ../nbs/04_docments.ipynb #6355b569
def _clean_text_sig(obj):
    if not (sig := getattr(obj, '__text_signature__', None)): return None
    sig = re.sub(r'\$\w+,?\s*', '', sig)
    return get_name(obj) + sig.replace('<unrepresentable>', '...')

# %% ../nbs/04_docments.ipynb #cfcc46bf
def _fmt_sig(name, params, ret_str, maxline, prefix='def'):
    "Format function signature with params and docment comments"
    lines,curr = [],[]
    for fmt,doc in params:
        comment = f' # {doc}' if doc else ''
        if curr and len(', '.join(curr))+len(fmt)+len(comment)>maxline:
            lines.append(', '.join(curr) + ',')
            curr = []
        curr.append(fmt)
        if doc: lines.append(', '.join(curr) + ',' + comment); curr = []
    if curr: lines.append(', '.join(curr))
    pstr = '\n    '.join(lines)
    return f"{prefix} {name}(\n    {pstr}\n{ret_str}"

# %% ../nbs/04_docments.ipynb #4550820f
def _fmt_default(o):
    if o is empty: return ''
    return o.__name__ if hasattr(o, '__name__') else repr(o)

class DocmentText(_DocmentBase):
    def __init__(self, obj, maxline=110, docstring=True):
        super().__init__(obj)
        self.maxline,self.docstring = maxline,docstring

    def _fmt_param(self, nm, p):
        anno,default = p.get('anno',empty), p.get('default',empty)
        return nm + (f':{_maybe_nm(anno)}' if anno != empty else '') + (f'={_fmt_default(default)}' if default != empty else '')
    
    @property
    def _ret_str(self):
        ret = self.dm.get('return', {})
        anno = f"->{_maybe_nm(ret.get('anno',empty))}" if ret.get('anno',empty) != empty else ''
        doc = f" # {ret['docment']}" if ret.get('docment') else ''
        return f"){anno}:{doc}"
    
    @property
    def params(self): return [(self._fmt_param(k,v), v.get('docment','')) for k,v in self.dm.items() if k != 'return']

    def __str__(self):
        prefix = 'async def' if inspect.iscoroutinefunction(self.obj) else 'def'
        if (sig := _clean_text_sig(self.obj)) and not self.params: sig_str = f"{prefix} {sig}"
        else: sig_str = _fmt_sig(get_name(self.obj), self.params, self._ret_str, self.maxline, prefix=prefix)
        docstr = f'    "{self.obj.__doc__}"' if self.docstring and self.obj.__doc__ else ''
        return f"{sig_str}\n{docstr}"
    
    __repr__ = __str__
    def _repr_markdown_(self): return f"```python\n{self}\n```"

# %% ../nbs/04_docments.ipynb #d44a7fe3
def sig2str(func, maxline=110):
    "Generate function signature with docments as comments"
    return DocmentText(func, maxline=maxline, docstring=False)

# %% ../nbs/04_docments.ipynb #07287425
def _docstring(sym):
    npdoc = parse_docstring(sym)
    return '\n\n'.join([npdoc['Summary'], npdoc['Extended']]).strip()

# %% ../nbs/04_docments.ipynb #9de89cb6
def _fullname(o):
    module,name = getattr(o, "__module__", None),qual_name(o)
    return name if module is None or module in ('__main__','builtins') else module + '.' + name

class ShowDocRenderer:
    def __init__(self, sym, name:str|None=None, title_level:int=3, maxline:int=110):
        "Show documentation for `sym`"
        if not hasattr(sym, '__signature__'): sym = getattr(sym, '__wrapped__', sym)
        sym = getattr(sym, 'fget', None) or getattr(sym, 'fset', None) or sym
        store_attr()
        self.nm = name or qual_name(sym)
        self.isfunc = inspect.isfunction(sym)
        try: self.sig = signature(sym, eval_str=True)
        except (ValueError,TypeError): self.sig = None
        self.docs = _docstring(sym)
        self.dm = DocmentText(sym, maxline=maxline, docstring=False)
        self.fn = _fullname(sym)

    __repr__ = basic_repr()

# %% ../nbs/04_docments.ipynb #e569d885
def _f_name(o): return f'<function {o.__name__}>' if isinstance(o, FunctionType) else None
def _fmt_anno(o): return inspect.formatannotation(o).strip("'").replace(' ','')

def _show_param(param):
    "Like `Parameter.__str__` except removes: quotes in annos, spaces, ids in reprs"
    kind,res,anno,default = param.kind,param._name,param._annotation,param._default
    kind = '*' if kind==inspect._VAR_POSITIONAL else '**' if kind==inspect._VAR_KEYWORD else ''
    res = kind+res
    if anno is not inspect._empty: res += f':{_f_name(anno) or _fmt_anno(anno)}'
    if default is not inspect._empty: res += f'={_f_name(default) or repr(default)}'
    return res

# %% ../nbs/04_docments.ipynb #59797eb7
def _ital_first(s:str):
    "Surround first line with * for markdown italics, preserving leading spaces"
    return re.sub(r'^(\s*)(.+)', r'\1*\2*', s, count=1)

# %% ../nbs/04_docments.ipynb #9184b175
class MarkdownRenderer(ShowDocRenderer):
    "Markdown renderer for `show_doc`"
    def _repr_markdown_(self):
        doc = f'```python\n\n{self.dm}\n\n```'
        if self.docs: doc += f"\n\n{_ital_first(self.docs)}"
        return doc

    def __repr__(self):
        doc = str(self.dm)
        if self.docs: doc += f'"""\n{self.docs}\n"""'
        return doc
    __str__=__repr__



def __getattr__(name):
     raise ImportError(
         f"Could not import '{name}' from fastcore.dispatch - this module has been moved to the fasttransform package.\n"
         "To migrate your code, please see the migration guide at: https://answerdotai.github.io/fasttransform/fastcore_migration_guide.html"
     )